from abc import ABC, abstractmethod
from ..constants import Secret

class Table(ABC):
    
    @abstractmethod
    def get_table(self)->object:
        "Returns the implementation of the DB table object"
        return self   
    
    @abstractmethod
    def backup_table(self, backup_name:str)->object:
        "Backs-up the table and returns an identifier of the backup"
        return None
    
    @abstractmethod
    def cleanup_table_backups(self, backup_name:str)->None:
        "Delete all table backups"
        pass

    @abstractmethod
    def drop_table(self)->None:
        "Drops the DB table"
        pass

    @abstractmethod
    def has_table(self, table_name:str=None)->bool:
        "Checks if the table exists"
        return False
    
    @abstractmethod
    def create_table(self, table_name:str=None)->None:
        "Creates a table if it does not exist"
        pass

    @abstractmethod
    def insert_record(self,
                      secret:Secret,
                      mem_pwd:str,
                      salt:str,
                      timestamp:str) -> None:
        "insert a record in the secrets table"
        pass

    @abstractmethod
    def insert_encrypted_record(self,
                                secret:Secret,
                                timestamp:str) -> None:
        "insert a pre-encrypted record in the secrets table"
        pass
    
    @abstractmethod
    def get_record(self,secret: Secret) -> Secret:
        "retrieves an encrypted record keyed by domain and access, as a dictionary"
        return None
        
    
    @abstractmethod
    def update_record_single_info(self,
                                  secret:Secret,
                                  mem_pwd:str,
                                  salt:str,
                                  timestamp:str) -> None:
        "update a record in the secrets table, with a maximum of one extra info stored in the Secret record"
        pass
    
    @abstractmethod
    def update_record_info_dictionary(self,
                                      secret: Secret,
                                      timestamp:str) -> None:
        "update the info dictionary for a record in the secrets table"
        pass
    
    @abstractmethod
    def delete_record(self, secret:Secret) -> None:
        "delete a single record based on domain and access keys"
        pass
    
    @abstractmethod
    def query_record(self, secret:Secret) -> list: 
        "return a list of secrets matching the domain key passed"
        return []
    
    


from password_strength import PasswordPolicy
from secretwallet.constants import PWD_LENGTH, PWD_NUMBERS, PWD_SPECIAL, PWD_UPPER, PWD_ATTEMPTS, parameters
from secretwallet.utils.logging import get_logger

import secretwallet.session.client as sc
import secretwallet.utils.ioutils as iou


logger = get_logger(__name__, parameters.get_log_level())
parameters.register_logger(__name__, logger)

__policy = PasswordPolicy.from_names(length  = PWD_LENGTH,
                                   uppercase = PWD_UPPER,
                                   numbers   = PWD_NUMBERS,
                                   special   = PWD_SPECIAL )
__explanation = {f'Length({PWD_LENGTH})'   : f'The minimum password length is {PWD_LENGTH}',
                 f'Uppercase({PWD_UPPER})' : f'The minimum number of upper-case characters in password is {PWD_UPPER}',
                 f'Numbers({PWD_NUMBERS})' : f'The minimum number of numbers in password is {PWD_NUMBERS}',
                 f'Special({PWD_SPECIAL})' : f'The minimum number of special characters in password is {PWD_SPECIAL}',}

def validate(pwd):
    return __policy.test(pwd)

def explain(pwd):
    faults = validate(pwd) #it is a set
    if len(faults)==0:
        return None
    else:
        return __explanation[str(faults[0])]

def get_password(prompt, attempts):
    num_attempts = attempts
    while num_attempts >0:
        print(f"*** {num_attempts} attempts left ***")
        p1 = iou.my_getpass(f"{prompt}-First entry  :")
        explanation = explain(p1)
        if explanation is not None:
            print(explanation)
            num_attempts -= 1
            continue
        p2 = iou.my_getpass(f"{prompt}-Verification :")
        if p1 != p2:
            message = "The two passwords are different, try again"
            print(message)
            logger.error(message)
            num_attempts -= 1
            continue
        return p1
    message = "Too many attempts at entering a valid password. Goodbye!"
    print(message)
    logger.error(message)
    exit(1)

def get_password_untested(prompt):
    return iou.my_getpass(f"{prompt} :")

def get_memorable_password(tested = False):
    """Get the memorable password, either from the live session or from the client prompt.
    input:
    tested    a boolean flag. If true then the password strength is ensured
              and the password is checked against a second entry
    output:
    return a pair of values (memorable, need_session) where memorable is the
    memorable password and need_session is a flag that tells if a new session needs to be started.
    """
    memorable = None
    if parameters.is_in_shell(): #in shell mode the password is kept in memory in the parameters
        memorable = parameters.get_memorable_pwd()
        if memorable is not None:
            return (memorable, False)
        else:
            if tested:
                memorable = get_password("Enter the memorable password", PWD_ATTEMPTS)
            else:
                memorable = get_password_untested("Enter the memorable password")
            parameters.set_memorable_pwd(memorable)

            return (memorable, False)
    elif sc.is_connected(): #in bash shell the password is kept in a separate process
        res = sc.get_session_password()
        if res[0] == 'fresh':
            return (res[1], False)
        else:
            if tested:
                memorable = get_password("Enter the memorable password", PWD_ATTEMPTS)
            else:
                memorable = get_password_untested("Enter the memorable password")
            sc.set_session_password(memorable)
            return (memorable, False)
    else: #session not running
        if tested:
            memorable = get_password("Enter the memorable password", PWD_ATTEMPTS)
        else:
            memorable = get_password_untested("Enter the memorable password")
        return (memorable,True)


import os
import json
import configparser
from secretwallet.constants import parameters, CONFIG_FILE, CREDENTIALS_FILE
from secretwallet.utils.fileutils import touch
from secretwallet.utils.cryptutils import encrypt_key
from secretwallet.utils.dbutils import create_table, has_table
from secretwallet.utils.password_manager import get_password

def has_configuration(config_file=CONFIG_FILE):
    """Checks if the configurations file CONFIG_FILE exists
    input:
    config_file    a path to the configuration file
    output:
    Boolean indicating if the configuration file exists"""
    return os.path.exists(config_file)

def has_credentials(credentials_file=CREDENTIALS_FILE):
    """Checks if the aws-credentials file CREDENIALS_FILE exists
    input:
    credentials_file   a path to the credentials file
    output:
    Boolean indicating if the configturation file exists"""
    return os.path.exists(credentials_file) 

def set_credentials(config_file, access_key, secret_access_key,region):
    """make or update credentials
    input:
    config_file       a path to the AWS configuration file
    access_key        the AWS access key
    secret_access_key the secret access key
    region            the AWS region
    """
    parser = configparser.ConfigParser()
    if has_credentials(config_file):
        parser.read(config_file)
    else:
        touch(config_file)        
    parser[parameters.get_profile_name()] = {'aws_access_key_id'    :  access_key,
                                             'aws_secret_access_key':  secret_access_key,
                                             'region'               :  region}

    with open(config_file, 'w') as f:
        parser.write(f)


def get_credentials(config_file=CREDENTIALS_FILE):
    """Retrieves AWS credentials
    input:
    config_file       a path to the AWS configuration file
    output:
    a dictionary with the AWS credentials
    """
    parser = configparser.ConfigParser()
    ret = {}
    if has_credentials(config_file):
        parser.read(config_file)
        ret = parser[parameters.get_profile_name()]
    return ret

def get_credentials_sections(config_file=CREDENTIALS_FILE):
    """Retrieves the AWS credentials sections
    input:
    config_file       a path to the AWS configuration file
    output:
    a list of AWS credentials sections
    """
    parser = configparser.ConfigParser()
    ret = []
    if has_credentials(config_file):
        parser.read(config_file)
        ret = parser.sections()
    return ret


def get_configuration(config_file = CONFIG_FILE):
    """Read the configuration file and returns it as a dictionary
    input:
    config_file    a path to the configuration file
    output:
    A data dictionary containing the configuration"""
    if not os.path.exists(config_file):
        raise FileNotFoundError("Missing configuration file: run the configuration script secret_wallet_conf")
    with open(config_file, 'r') as cfile:
        return json.load(cfile)
    
def list_configuration(config_file = CONFIG_FILE):
    """List the configuration parameters
    input:
    config_file    a path to the configuration file
    """
    conf = get_configuration(config_file)
    display_configuration(config_file, 'secret wallet configuration is located', conf)
    
def set_configuration_data(conf_data, conf_file = CONFIG_FILE):
    with open(conf_file, 'w') as cfile:
        json.dump(conf_data, cfile)    
    
def set_configuration(conf_key, profile = None, table = None, salt = None, config_file = CONFIG_FILE):
    """This writes the system configuration file with the specified overrides and the encrypted salt
       If the configuration file exists, this function returns an error message, since reconfiguing the salt
       requires changes to all the encripted information in the remote DB.
       If the configuration already exists it should raise an error
       input:
       conf_key        the encrypted configuration key
       profile          the override of the default profile (optionall)
       table            the override of the default table name  (optional)
       salt             the override of the default pre-salt string
       config_file      the configuration file, defaults to fixed location in CONFIG_FILE
       """
    if has_configuration(config_file) and has_table(table):
        raise RuntimeError("Found pre-existing configuration in %s. To reconfigure the secretes use the reconf command"%config_file)
    
    conf = {'key': conf_key}
    if profile is not None:
        conf['profile'] = profile
    if table is not None:
        conf['table_name'] = table
    if salt is not None:
        conf['salt'] = salt
    os.makedirs(os.path.dirname(config_file), exist_ok=True)
    set_configuration_data(conf, config_file)
    
def load_configurations(conf_file = CONFIG_FILE, credentials_file = CREDENTIALS_FILE):
    """
    Loads the configuration and aws credentials
    """
    try:
        if not has_credentials(credentials_file):
            raise FileNotFoundError("Missing configuration file: run the configuration script secret_wallet_conf")    
        
        if not has_configuration(conf_file):
            raise FileNotFoundError("Missing configuration file: run the configuration script secret_wallet_conf")
        parameters.set_data(get_configuration(conf_file))
    except Exception as e:
        print(e)
        exit(1)
        
def display_configuration(conf_file, content, conf):
    print(f"\nThe {content} at {conf_file}")
    for k,v in conf.items():
        if k=='key': continue
        print(f'{k:30} = {v:<40}')
        
        
def make_configurations():
    "Main configuration script"
    print("\nMain configuration script for your secret wallet.")
    print("Please press return to accept the pre-set values in square brackets, or type a new value:\n")
    
    answ = input("\nDo you want to configure the AWS credentials? (yes|skip) ")
    if answ.lower().startswith('y'):    
        profile           = input('{0:30}[{1:>30}] = '.format('AWS profile name',parameters.get_profile_name()))
        if len(profile) == 0:
            profile = parameters.get_profile_name()
        if has_credentials() and profile in get_credentials_sections():
            print('The AWS profile {0} is already in use. Choose another or reconfigure'.format(profile))
            exit(1)
        parameters.set_profile_name(profile)
        access_key        = input('{0:30}[{1:>30}] = '.format('AWS access key id',''))
        secret_access_key = input('{0:30}[{1:>30}] = '.format('AWS secret access key',''))
        region            = input('{0:30}[{1:>30}] = '.format('AWS region',''))
        cred = {'aws access key id'     : access_key,
                'aws seceret access key': secret_access_key,
                'aws region'            : region}
        
        display_configuration(CREDENTIALS_FILE, 'AWS credentials are located',cred)
        answ = input("\nDo you want to set the credentials? (yes|skip|exit) ")
        if answ.lower().startswith('y'):
            set_credentials(CREDENTIALS_FILE, access_key, secret_access_key, region)
        elif answ.lower().startswith('s'):
            pass
        else:
            exit(1)
        
    answ = input("\nDo you want to configure the the secret-wallet parameters? (yes|skip) ")
    if answ.lower().startswith('y'):
        profile           = input('{0:30}[{1:>30}] = '.format('AWS profile name',parameters.get_profile_name()))
        if len(profile) == 0:
            profile = parameters.get_profile_name()        
        table             = input('{0:30}[{1:>30}] = '.format('DynameDB table name',parameters.get_table_name()))
        if len(table) == 0:
            table = parameters.get_table_name()                
        if has_configuration() and has_table(table):
            
            print('The secret-wallet has been configured previously for the same table.\nTo protect secrets, you need to call a reconfigure procedure')
            exit(1)    
        conf_pwd = get_password('Configuration password', 6)
        conf_key = encrypt_key(conf_pwd)
            
        conf = {'configuration key': conf_key,
                'profile': profile,
                'table': table}
        parameters.set_profile_name(profile)
        parameters.set_table_name(table)
        display_configuration(CONFIG_FILE, 'secret wallet configuration is located', conf)     
        answ = input("\nDo you want to set the configuration parameters? (yes|exit) ")
        if answ.lower().startswith('y'):
            try:
                set_configuration(conf_key, profile, table, None, CONFIG_FILE)
                create_table(table)
            except Exception as e:
                print(e)
                print("Could not write the configuration file. Make sure you have AWS connection and try again")
        else:
            exit(1)
            
    answ = input("\nDo you want to select or change the storage type: local or remote? (yes|skip) ")
    if answ.lower().startswith('y'):
        while (True):
            storage = input('{0:30}[{1:>30}] = '.format('storage type (aws_dynamo|local_sqlite)',parameters.get_storage_type()))
            if len(storage) == 0:
                storage = parameters.get_storage_type()
            if storage.lower() in ('aws_dynamo','local_sqlite'):
                break  
        conf = get_configuration(CONFIG_FILE)
        conf["storage_type"] = storage
        display_configuration(CONFIG_FILE, 'secret wallet configuration is located', conf)     
        answ = input("\nDo you want to set the configuration parameters? (yes|exit) ")
        if answ.lower().startswith('y'):
            try:
                set_configuration_data(conf, CONFIG_FILE)
            except Exception as e:
                print(e)
                print("Could not write the configuration file. Make sure you have AWS connection and try again")
        else:
            exit(1)
        
    



# This is added to faciliting testing without installing secretwallet
# the executable can be started by
# python -m secretwallet <add here the arguments that you need>

from .main import main, configure

main()
#configure()


from multiprocessing.connection import Client

from secretwallet.constants import parameters
from secretwallet.utils.logging import get_logger
from secretwallet.utils import is_posix


logger = get_logger(__name__, parameters.get_log_level())
parameters.register_logger(__name__, logger)

def get_session_password():
    if not is_posix():
        raise NotImplementedError("Client/Server daemons not supported on this system")
    logger.debug("Retrieving session password")
    conn = Client(parameters.get_session_address(), authkey=parameters.get_session_connection_password())
    conn.send({'action':'get','password':None})
    r = conn.recv()
    conn.close()
    return r['status'],r['password']

def ping_session_server():
    if not is_posix():
        raise NotImplementedError("Client/Server daemons not supported on this system")
    logger.debug("Pinging session server")
    conn = Client(parameters.get_session_address(), authkey=parameters.get_session_connection_password())
    conn.send({'action':'ping','password':None})
    r = conn.recv()
    conn.close()
    return r['status']

def set_session_password(pwd):
    if not is_posix():
        return
    logger.debug("Setting session password")
    conn = Client(parameters.get_session_address(), authkey=parameters.get_session_connection_password())
    conn.send({'action':'set','password':pwd})
    r = conn.recv()
    conn.close()
    logger.debug(r['status'])

def stop_service():
    if not is_posix():
        return
    logger.debug("Stopping the service")
    if is_connected():
        conn = Client(parameters.get_session_address(), authkey=parameters.get_session_connection_password())
        conn.send({'action':'stop','password':None})
        r = conn.recv()
        conn.close()
        logger.debug(r['status'])

def is_connected():
    if parameters.is_in_shell():
        return False #the secret_wallet shell has a different way to keep the password
    if not is_posix():
        return False  #only posix systems support daemons in the way required
    try:
        return ping_session_server()=="OK"
    except:
        return False






from multiprocessing import Process
from time import sleep

import pytest
from secretwallet.session.client import get_session_password, set_session_password, stop_service, is_connected, ping_session_server
from secretwallet.session.service import my_session

import psutil

def kill_all_procs():
    pname = 'secret_wallet'
    for proc in psutil.process_iter():
        if proc.name() == pname:
            proc.kill()


@pytest.fixture
def set_up():
    kill_all_procs()
    lifetime = 12
    timeout = 6
    value = 'message'

    #kill the listener
    if is_connected():
        stop_service()
    #now start the services
    p =Process(target=my_session, args =(value, lifetime, timeout))
    p.start()
    while not is_connected():
        pass

    yield

    p.terminate()


def test_get_password(set_up):
    assert 'message' == get_session_password()[1]

def test_set_password(set_up):
    res = get_session_password()
    assert 'fresh'   == res[0]
    assert 'message' == res[1]
    set_session_password('nuova')
    res = get_session_password()
    assert 'fresh'   == res[0]
    assert 'nuova' == res[1]

def test_expiry_password(set_up):
    res = get_session_password()
    assert 'fresh'   == res[0]
    assert 'message' == res[1]
    sleep(6)
    res = get_session_password()
    assert 'stale'   == res[0]
    assert res[1] is None

def test_connection_status(set_up):
    assert is_connected()
    stop_service()
    assert not is_connected()

def test_lifetime(set_up):
    assert is_connected()
    sleep(12)
    assert not is_connected()

def test_ping(set_up):
    assert "OK"==ping_session_server()



import sys
from secretwallet.main import myparser
from secretwallet.main.configuration import load_configurations, make_configurations

def main():
    load_configurations()
    myparser.Parser()

def configure():
    make_configurations()
    
def shell():
    load_configurations()
    sys.argv=['secret_wallet','shell']
    myparser.Parser()
    



from contextlib import redirect_stdout
import io
import json
from multiprocessing import Process
import os
import sys
from time import sleep

import pytest
from secretwallet.constants import parameters
from secretwallet.main.myparser import Parser
from secretwallet.session.client import is_connected, stop_service
from secretwallet.session.service import my_session
from   secretwallet.utils.cryptutils import decrypt

import secretwallet.utils.cryptutils as cu
import secretwallet.utils.dbutils as du
import secretwallet.utils.ioutils as iou


DOMAIN = 'test_domain'
ACCESS = 'test_access'
UID    = 'me@HOME'
PWD    = 'pass'
INFO   = {'key':'value'}
MEM    = 'memorable'
LIFETIME = 240
TIMEOUT = 60


old_input = iou.my_input
old_output = iou.my_output
old_getpass = iou.my_getpass


@pytest.fixture
def cli_test_set_up():
    #mocking the user input
    iou.my_input  = lambda _:'yes'
    iou.my_output = lambda message, with_exit=False, with_logging=False: print(message)

    if is_connected():
        stop_service()
    p =Process(target=my_session, args =(MEM, LIFETIME, TIMEOUT))
    p.start()
    while not is_connected():
        pass

    du.create_table(parameters.get_table_name())
    du.insert_secret(DOMAIN, ACCESS, UID, PWD, INFO, MEM)


    yield

    iou.my_input = old_input
    iou.my_output = old_output
    iou.my_getpass = old_getpass
    du.delete_secret(DOMAIN,ACCESS)

    if is_connected():
        stop_service()
    p.terminate()


@pytest.mark.integration
def test_help(cli_test_set_up):
    sys.argv=['secret_wallet','help']
    #output redirection to string
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert 'list of secretwallet commands' in buf.getvalue()


@pytest.mark.integration
def test_list(cli_test_set_up):
    sys.argv=['secret_wallet','list']
    #output redirection to string
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "<domain>" in buf.getvalue()

@pytest.mark.integration
def test_empty_list(cli_test_set_up):
    sys.argv=['secret_wallet','list','-d','xxx']
    #output redirection to string
    with io.StringIO() as buf, redirect_stdout(buf):
        try:
            Parser()
        except:
            assert False, "An empty list should not raise and exception when formatted"
        assert "****" in buf.getvalue()


@pytest.mark.integration
def test_list_domain(cli_test_set_up):
    sys.argv=['secret_wallet','list','-d',DOMAIN]
    #output redirection to string
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert ACCESS in buf.getvalue()


@pytest.mark.integration
def test_set_secret(cli_test_set_up):
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', 'test_access_2', '-u','x@y','-p','mamma']
    #output redirection to string
    try:
        with io.StringIO() as buf, redirect_stdout(buf):
            Parser()
            du.list_secrets(DOMAIN)
            assert 'test_access_2' in buf.getvalue()
    finally:
        du.delete_secret(DOMAIN,'test_access_2')

@pytest.mark.integration
def test_get_secret(cli_test_set_up):
    sys.argv=['secret_wallet','get','-d',DOMAIN, '-a', ACCESS]
    #output redirection to string
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert ACCESS in buf.getvalue()

@pytest.mark.integration
def test_update_info(cli_test_set_up):
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', ACCESS, '-ik','first_key','-iv','first_value']
    Parser()
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', ACCESS, '-ik','second_key','-iv','second_value']
    Parser()
    res = du.get_secret(DOMAIN, ACCESS, 'memorable', parameters.get_salt_key())
    assert 3 == len(res.info)
    assert 'value' == res.info['key']
    assert 'first_value' == res.info['first_key']
    assert 'second_value' == res.info['second_key']

@pytest.mark.integration
def test_rename_secret(cli_test_set_up):
    new_domain = "new domain_01"
    new_access = "new_access_01"

    #delete first
    du.delete_secret(DOMAIN, ACCESS)
    #then set
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', ACCESS, '-ik','first_key','-iv','first_value']
    Parser()
    assert du.has_secret(DOMAIN, ACCESS)
    assert not du.has_secret(new_domain, new_access)

    #now rename
    sys.argv=['secret_wallet','rename','-d',DOMAIN, '-a', ACCESS, '-nd', new_domain,'-na', new_access]
    Parser()
    assert not du.has_secret(DOMAIN, ACCESS)
    assert du.has_secret(new_domain, new_access)
    du.delete_secret(new_domain, new_access)

@pytest.mark.integration
def test_query_by_domain(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1]
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2]
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now querying by domain
    sys.argv=['secret_wallet','query','-d','pera']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "cotta" in buf.getvalue()
        assert "bella" not in buf.getvalue()

@pytest.mark.integration
def test_query_by_access(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1]
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2]
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now querying by domain
    sys.argv=['secret_wallet','query','-a','pera']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "cotta" not in buf.getvalue()
        assert "bella" in buf.getvalue()

@pytest.mark.integration
def test_query_by_pattern(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1]
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2]
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now querying by domain
    sys.argv=['secret_wallet','query','pera']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "cotta" in buf.getvalue()
        assert "bella" in buf.getvalue()


@pytest.mark.integration
def test_qget_first_secret(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #when the list of secrete is returned they are ordered alphabetically
    #by domain, access. Hence the first secret is bella,pera, or the second record here
    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now running a qget command with some mockable input
    iou.my_input = iou.MockableInput(["1"])
    sys.argv=['secret_wallet','qget','pera']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "first record" in buf.getvalue() #get the inner value in the secret
        assert "second record" not in buf.getvalue()


@pytest.mark.integration
def test_qget_second_secret(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #when the list of secrete is returned they are ordered alphabetically
    #by domain, access. Hence the first secret is bella,pera, or the second record here
    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now running a qget command with some mockable input
    iou.my_input = iou.MockableInput(["2"])
    sys.argv=['secret_wallet','qget','pera']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "first record" not in buf.getvalue() #get the inner value in the secret
        assert "second record" in buf.getvalue()

@pytest.mark.integration
def test_qget_wrong_input(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #when the list of secrete is returned they are ordered alphabetically
    #by domain, access. Hence the first secret is bella,pera, or the second record here

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now running a qget command with some mockable input
    iou.my_input = iou.MockableInput(["string"])
    sys.argv=['secret_wallet','qget','pera']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "first record" not in buf.getvalue() #get the inner value in the secret
        assert "second record" not in buf.getvalue()
        assert "I need a number"  in buf.getvalue()

@pytest.mark.integration
def test_delete_info_item(cli_test_set_up):
    domain = 'pera'
    access = "cotta"
    key1 = 'bella'
    value1 = 'pupa'
    key2 = 'toste'
    value2 = 'mele'

    #delete first
    du.delete_secret(domain, access)
    #then set
    sys.argv=['secret_wallet','set','-d',domain, '-a', access, '-ik', key1, '-iv', value1]
    Parser()
    sys.argv=['secret_wallet','set','-d',domain, '-a', access, '-ik', key2, '-iv',value2]
    Parser()

    assert du.has_secret(domain, access)

    sys.argv=['secret_wallet','get', '-d', domain, '-a', access]
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "pupa" in buf.getvalue()
        assert "mele" in buf.getvalue()

    #now remove one item from the dictionary
    sys.argv=['secret_wallet','delete', '-d', domain, '-a', access, '-ik', key2]
    Parser()

    #check that the item is gone
    sys.argv=['secret_wallet','get', '-d', domain, '-a', access]
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "pupa" in buf.getvalue()
        assert "mele" not in buf.getvalue()

@pytest.mark.integration
def test_rename_secret_no_new_values(cli_test_set_up):
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', ACCESS, '-ik','first_key','-iv','first_value']
    Parser()
    sys.argv=['secret_wallet','rename','-d',DOMAIN, '-a', ACCESS]
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "No new keys have been passed" in buf.getvalue()

@pytest.mark.integration
def test_rename_secret_same_values(cli_test_set_up):
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', ACCESS, '-ik','first_key','-iv','first_value']
    Parser()
    sys.argv=['secret_wallet','rename','-d',DOMAIN, '-a', ACCESS, '-nd', DOMAIN, '-na', ACCESS]
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "Both new values are the same as the originals: nothing to do" in buf.getvalue()

@pytest.mark.integration
def test_rename_secret_wrong_values(cli_test_set_up):
    sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', ACCESS, '-ik','first_key','-iv','first_value']
    Parser()
    sys.argv=['secret_wallet','rename','-d',DOMAIN, '-a', 'wrong', '-nd', DOMAIN, '-na', ACCESS]
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "Could not find the secret to rename" in buf.getvalue()

@pytest.mark.integration
def test_conf_list(cli_test_set_up):
    sys.argv=['secret_wallet','conf','-l']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "secret wallet configuration is located" in buf.getvalue()

@pytest.mark.integration
def test_wrong_salt(cli_test_set_up):
    my_access = 'another'
    other_key = cu.encrypt_key('pirillo')
    du.insert_secret(DOMAIN, my_access, 'login', 'password', None, 'memorable', other_key)
    sleep(1)
    #the following shoud produce and InvalidToken error
    sys.argv=['secret_wallet','get','-d',DOMAIN, '-a', my_access]
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert 'InvalidToken' in buf.getvalue()

@pytest.mark.integration
def test_wrong_memorable_password(cli_test_set_up):
    my_access = 'another'
    try:
        #insert
        sys.argv=['secret_wallet','set','-d',DOMAIN, '-a', my_access, '-u','login','-p','password']
        Parser()
        #now change the memorable in the session
        sys.argv=['secret_wallet','client','-a','set','-v','azzo']
        Parser()
        sleep(1)
        #the following shoud produce and InvalidToken error
        sys.argv=['secret_wallet','get','-d',DOMAIN, '-a', my_access]
        with io.StringIO() as buf, redirect_stdout(buf):
            Parser()
            assert 'InvalidToken' in buf.getvalue()
    finally:
        du.delete_secret(DOMAIN,my_access)

@pytest.mark.integration
def test_shell_set_help(cli_test_set_up):
    #mocking input to pass a 'set -h' command in a shell
    iou.my_input = iou.MockableInput(['set -h','quit'])
    sys.argv=['secret_wallet','shell']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert 'usage: secret_wallet set' in buf.getvalue()

@pytest.mark.integration
def test_shell_set_get_delete(cli_test_set_up):
    password = 'Arz12@gh67!caz'
    #mocking password retrieval
    iou.my_getpass = lambda _: password
    #mocking input to pass a 'set ...' command in a shell
    iou.my_input = iou.MockableInput(["set -d shell_test -a test -ik test -iv 'this is a test'",
                                      'get -d shell_test -a test',
                                      'delete -d shell_test -a test',
                                      'yes',
                                      'quit'])
    sys.argv=['secret_wallet','shell']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert 'this is a test' in buf.getvalue()

    #now check it is not there any longer
    assert not du.has_secret('shell_test','test')

@pytest.mark.integration
def test_shell_set_rename_get_delete(cli_test_set_up):
    password = 'Arz12@gh67!caz'
    #mocking password retrieval
    iou.my_getpass = lambda _: password
    #mocking input to pass a 'set ...' command in a shell
    iou.my_input = iou.MockableInput(["set -d shell_test -a test -ik test -iv 'this is a test'",
                                      'rename -d shell_test -a test -na test2',
                                      'yes',
                                      'get -d shell_test -a test2',
                                      'delete -d shell_test -a test2',
                                      'yes',
                                      'quit'])
    sys.argv=['secret_wallet','shell']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert 'this is a test' in buf.getvalue()

    #now check it is not there any longer
    assert not du.has_secret('shell_test','test')
    assert not du.has_secret('shell_test','test2')

@pytest.mark.integration
def test_dump(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    sys.argv=['secret_wallet','dump']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "first record" in buf.getvalue() #get the inner value in the secret
        assert "second record" in buf.getvalue()

@pytest.mark.integration
def test_dump_to_file(cli_test_set_up):
    path = "test_dump_tp_file.txt"

    # first clean up the file if it exists
    if os.path.exists(path):
        os.remove(path)

    #the set the data
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now do the test
    try:
        sys.argv=['secret_wallet','dump','-f', path]
        Parser()

        #now assert the file exists
        assert os.path.exists(path)

        #and that it contains the data
        with open(path) as f:
            data = f.read()
            assert "first record" in data
            assert "second record" in data

    finally:
        os.remove(path)

@pytest.mark.integration
def test_save_decrypted(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    sys.argv=['secret_wallet','save','-d']
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert "first record" in buf.getvalue() #get the inner value in the secret
        assert "second record" in buf.getvalue()

@pytest.mark.integration
def test_save_to_file_decrypted(cli_test_set_up):
    path = "test_dump_to_file.json"

    # first clean up the file if it exists
    if os.path.exists(path):
        os.remove(path)

    #the set the data
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','first record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','second record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now do the test
    try:
        sys.argv=['secret_wallet','save','-f', path,'-d']
        Parser()

        #now assert the file exists
        assert os.path.exists(path)

        #and that it contains the data
        with open(path) as f:
            data = json.load(f)
            first_record = [x for x in data if x["domain"]==domain1 and x["access"]==access1][0]
            second_record = [x for x in data if x["domain"]==domain2 and x["access"]==access2][0]
            assert "first record" == first_record["info"]["idx"]
            assert "second record" == second_record["info"]["idx"]

    finally:
        os.remove(path)

@pytest.mark.integration
def test_load_from_file_decrypted(cli_test_set_up):
    path = "test_backup_file.json"

    # first clean up the file if it exists
    if os.path.exists(path):
        os.remove(path)

    #all the pre-existing secrets
    secrets = du.list_secrets(None)
    #clean-up
    du.delete_secrets(secrets)
    secrets = du.list_secrets(None)
    assert 0 == len(secrets)

    #now add some secrets

    #set the data
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #set the secrets
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','first record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','second record']
    Parser()

    #check the status now
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)
    assert 2 == len(du.list_secrets(None)) #there should be two secrets
    s1a = du.get_secret(domain1, access1, MEM)
    s2a = du.get_secret(domain2, access2, MEM)

    #now save to a file
    try:
        sys.argv=['secret_wallet','save','-f', path,'-d']
        Parser()

        #now assert the file exists
        assert os.path.exists(path)

        #now clean-up the secrets again
        du.delete_secrets(du.list_secrets(None))
        assert 0 == len(du.list_secrets(None))

        #and reload from file
        sys.argv=['secret_wallet','load','-f', path, '-d']
        Parser()

        assert du.has_secret(domain1, access1)
        assert du.has_secret(domain2, access2)
        assert 2 == len(du.list_secrets(None)) #there should be two secrets  again
        s1b = du.get_secret(domain1, access1, MEM)
        s2b = du.get_secret(domain2, access2, MEM)
        assert s1a == s1b
        assert s2a == s2b

        #now try to reload the file without cleaning the table,
        #it should still work without side effects
        sys.argv=['secret_wallet','load','-f', path,'-d']
        Parser()

        assert du.has_secret(domain1, access1)
        assert du.has_secret(domain2, access2)
        assert 2 == len(du.list_secrets(None)) #there should be two secrets  again
        s1b = du.get_secret(domain1, access1, MEM)
        s2b = du.get_secret(domain2, access2, MEM)
        assert s1a == s1b
        assert s2a == s2b

    finally:
        os.remove(path)

@pytest.mark.integration
def test_save_encrypted(cli_test_set_up):
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','second record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','first record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    sys.argv=['secret_wallet','save'] #without the -d option it is encrypted
    with io.StringIO() as buf, redirect_stdout(buf):
        Parser()
        assert 'pera' in buf.getvalue() #this should be in clear
        assert 'idx' in buf.getvalue() #this should be in clear
        assert "first record" not in buf.getvalue() #this should be encrypted
        assert "second record" not in buf.getvalue()
        
@pytest.mark.integration
def test_save_to_file_encrypted(cli_test_set_up):
    path = "test_dump_to_file.json"

    # first clean up the file if it exists
    if os.path.exists(path):
        os.remove(path)

    #the set the data
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #delete first
    du.delete_secret(domain1, access1)
    du.delete_secret(domain2, access2)
    #then set
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','first record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','second record']
    Parser()
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)

    #now do the test
    try:
        sys.argv=['secret_wallet','save','-f', path]
        Parser()

        #now assert the file exists
        assert os.path.exists(path)

        #and that it contains the data
        with open(path) as f:
            data = json.load(f)
            first_record = [x for x in data if x["domain"]==domain1 and x["access"]==access1][0]
            second_record = [x for x in data if x["domain"]==domain2 and x["access"]==access2][0]
            assert "first record"  == decrypt(first_record["info"]["idx"],MEM,parameters.get_salt_key())
            assert "second record" == decrypt(second_record["info"]["idx"],MEM,parameters.get_salt_key())

    finally:
        os.remove(path)
        
@pytest.mark.integration
def test_load_from_file_encrypted(cli_test_set_up):
    path = "test_backup_file.json"

    # first clean up the file if it exists
    if os.path.exists(path):
        os.remove(path)

    #all the pre-existing secrets
    secrets = du.list_secrets(None)
    #clean-up
    du.delete_secrets(secrets)
    secrets = du.list_secrets(None)
    assert 0 == len(secrets)

    #now add some secrets

    #set the data
    domain1 = 'pera'
    access1 = "cotta"
    domain2 = 'bella'
    access2 = 'pera'

    #set the secrets
    sys.argv=['secret_wallet','set','-d',domain1, '-a', access1, '-ik', 'idx', '-iv','first record']
    Parser()
    sys.argv=['secret_wallet','set','-d',domain2, '-a', access2, '-ik', 'idx', '-iv','second record']
    Parser()

    #check the status now
    assert du.has_secret(domain1, access1)
    assert du.has_secret(domain2, access2)
    assert 2 == len(du.list_secrets(None)) #there should be two secrets
    s1a = du.get_secret(domain1, access1, MEM)
    s2a = du.get_secret(domain2, access2, MEM)

    #now save to a file
    try:
        sys.argv=['secret_wallet','save','-f', path]
        Parser()

        #now assert the file exists
        assert os.path.exists(path)

        #now clean-up the secrets again
        du.delete_secrets(du.list_secrets(None))
        assert 0 == len(du.list_secrets(None))

        #and reload from the encrypted file
        sys.argv=['secret_wallet','load','-f', path]
        Parser()

        assert du.has_secret(domain1, access1)
        assert du.has_secret(domain2, access2)
        assert 2 == len(du.list_secrets(None)) #there should be two secrets  again
        s1b = du.get_secret(domain1, access1, MEM)
        s2b = du.get_secret(domain2, access2, MEM)
        assert s1a == s1b
        assert s2a == s2b

        #now try to reload the file without cleaning the table,
        #it should still work without side effects
        sys.argv=['secret_wallet','load','-f', path]
        Parser()

        assert du.has_secret(domain1, access1)
        assert du.has_secret(domain2, access2)
        assert 2 == len(du.list_secrets(None)) #there should be two secrets  again
        s1b = du.get_secret(domain1, access1, MEM)
        s2b = du.get_secret(domain2, access2, MEM)
        assert s1a == s1b
        assert s2a == s2b

    finally:
        os.remove(path)
        


'''
Created on 16 Dec 2019

@author: gualtiero
'''

import pytest
import os
import secretwallet.utils.cryptutils as cu 
from secretwallet.main.configuration import set_configuration, get_configuration
from secretwallet.utils.cryptutils import encrypt_key
from secretwallet.constants import parameters

@pytest.fixture
def set_up():
    old_data = parameters.get_data()
    path = os.path.dirname(__file__)
    conf_file = os.path.join(path,'data','.secretwallet','test.json')
    yield conf_file
    
    if os.path.exists(conf_file):
        os.remove(conf_file)
        os.rmdir(os.path.dirname(conf_file))
    #reset the original parameters
    parameters.set_data(old_data)
    
        
def test_get_encrypted_key():
    passwd = u"passwd"
    ekey = cu.encrypt_key(passwd)
    assert ekey is not None 
    assert isinstance(ekey, str) 
    
    ekey2 = cu.encrypt_key(passwd)
    assert ekey2 == ekey    
            
def test_configure_save_key(set_up):
    conf_file = set_up
    passwd = u"passwd"
    set_configuration(encrypt_key(passwd), None, None, None, conf_file)
    
    data = get_configuration(conf_file)
    key = data['key'].encode('latin1')
    assert key is not None 
    assert isinstance(key, bytes)
    
def test_configure_save_many(set_up):
    conf_file = set_up
    passwd = u"passwd"
    profile = 'my profile'
    salt = 'my salt'
    table = 'my table'
    set_configuration(encrypt_key(passwd), profile, table, salt, conf_file)
    
    data = get_configuration(conf_file)
    key = data['key'].encode('latin1')
    assert key is not None 
    assert isinstance(key, bytes)
    assert profile == data['profile']
    assert salt == data['salt']
    assert table == data['table_name']    
    
def test_encrypt_decrypt_secret(set_up):
    conf_file = set_up
    c_pwd = u"passwd"
    m_pwd = u"memorabile"
    secret = u"mamma"
    set_configuration(encrypt_key(c_pwd), None, None, None, conf_file)
    parameters.set_data(get_configuration(conf_file))
    
    esecret = cu.encrypt(secret, m_pwd, parameters.get_salt_key())
    v1 = cu.decrypt(esecret, m_pwd, parameters.get_salt_key())
    esecret2 = cu.encrypt(secret, m_pwd, parameters.get_salt_key())
    v2 = cu.decrypt(esecret2, m_pwd, parameters.get_salt_key())
    assert v1 == v2
    assert v1 == secret
    
def test_encrypt_decrypt_no_config():
    c_pwd = u"passwd"
    m_pwd = u"memorabile"
    secret = u"mamma"
    key = cu.encrypt_key(c_pwd)    
    esecret = cu.encrypt(secret, m_pwd, key)
    v1 = cu.decrypt(esecret, m_pwd, key)
    esecret2 = cu.encrypt(secret, m_pwd, key)
    v2 = cu.decrypt(esecret2, m_pwd, key)
    assert v1 == v2
    assert v1 == secret
    
def test_encrypt_decrypt_info():
    c_pwd = u"passwd"
    m_pwd = u"memorabile"
    key = cu.encrypt_key(c_pwd)

    info={'first':'value_1','second':'value_2'}
    ien = cu.encrypt_info(info, m_pwd, key)
    ide = cu.decrypt_info(ien, m_pwd, key)
    for key in info:
        assert ide[key] == info[key]


import logging
from logging.handlers import RotatingFileHandler

from secretwallet.constants import LOG_FILE, LOG_MAX_FILE_SIZE, LOG_BACKUP_COUNT, make_log_level
from secretwallet.utils.fileutils import touch



def get_logger(name, log_level="info"):
    """Returns a logger
    input:
    name        the name of the logger, typically a python_module name
    level       the logging level
    """
    level = make_log_level(log_level)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    # Create the rotating file handler. Limit the size to 1000000Bytes ~ 1MB .
    touch(LOG_FILE) #if it does not exits, create it
    handler = RotatingFileHandler(LOG_FILE,
                                  mode='a',
                                  maxBytes=LOG_MAX_FILE_SIZE,
                                  backupCount=LOG_BACKUP_COUNT,
                                  encoding='utf-8',
                                  delay=0)
    handler.setLevel(level)
    # Create a formatter.
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # Add handler and formatter.
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


from datetime import datetime

from ..constants import Secret, CONFIG_FOLDER
from .table import Table
from os import path, system
from ..utils import is_posix
import sqlite3
from ..utils.cryptutils import encrypt, encrypt_info
import json
from pickle import NONE


"""
    Proxy class for a table on a local sqlite3 DB
    The methods can throw so the client (dbutils functions) need to capture them
"""
class LocalSqLiteTable(Table):
    
    def __init__(self, table_name):
        """
        Construct a sqlite table identified by a table name
        """
        self.table_name = table_name
        self.db_path = path.join(CONFIG_FOLDER,'secretwallet.sqlite')

                
    def get_table(self):
        "This is not used in sqlite, i.e. we don't use a table object"
        return Table.get_table(self)


    def backup_table(self, backup_name:str):
        backup_db_path = path.join(CONFIG_FOLDER,backup_name,'.sqlite')
        if is_posix(): #on linux systems
            system(f'cp {self.db_path} {backup_db_path}')
        else: #windows
            system(f'copy {self.db_path} {backup_db_path}')
        return backup_db_path


    def cleanup_table_backups(self, backup_name:str):
        backup_db_path = path.join(CONFIG_FOLDER,backup_name,'.sqlite')
        if is_posix(): #on linux systems
            system(f'rm {backup_db_path}')
        else: #windows
            system(f'del {backup_db_path}')


    def drop_table(self):
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute("DROP TABLE IF EXISTS ?;", (self.table_name,))
        finally:
            cursor.close()
            connection.close


    def has_table(self, table_name:str=None):
        "Checks if the table exists"
        result = False
        if table_name is None:
            table_name = self.table_name        
        query = """SELECT count(1)
                   FROM sqlite_schema 
                   WHERE type='table' 
                   AND name= ?;"""        
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,(table_name,))
            result = (cursor.fetchone()[0] == 1)
        finally:
            cursor.close()
            connection.close
        return result


    def create_table(self, table_name:str=None):
        "Creates a table if it does not exist"
        if table_name is None:
                table_name = self.table_name
        if self.has_table(table_name):
            return
        #todo: info field should be json        
        query = f"""CREATE TABLE IF NOT EXISTS [{self.table_name}] (
                      domain TEXT NOT NULL,
                      access TEXT NOT NULL,
                      uid TEXT,
                      pwd TEXT,
                      info TEXT, 
                      timestamp TEXT,
                      PRIMARY KEY (domain, access) );
                """
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query)
            connection.commit()
        finally:
            cursor.close()
            connection.close


    def insert_record(self, 
                     secret:Secret, 
                     mem_pwd:str, 
                     salt:str, 
                     timestamp:str=datetime.now().isoformat()):
        "insert a record in the secrets table"
        euid = encrypt(secret.user_id, mem_pwd, salt)
        epwd = encrypt(secret.password, mem_pwd, salt)
        einfo = json.dumps(encrypt_info(secret.info, mem_pwd, salt))
        
        record = (secret.domain,
                  secret.access,
                  euid,
                  epwd,
                  einfo,
                  timestamp)
        query = f"""INSERT INTO [{self.table_name}] 
                    VALUES (?1,?2,?3,?4,?5,?6) 
                    ON CONFLICT (domain, access)
                    DO UPDATE
                    SET uid = ?3,
                        pwd = ?4,
                        info = ?5,
                        timestamp = ?6;                    
                """
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,record)
            connection.commit()
        finally:
            cursor.close()
            connection.close        


    def insert_encrypted_record(self, 
                                secret:Secret, 
                                timestamp:str=datetime.now().isoformat()):
        "insert a pre-encrypted record in the secrets table"        
        record = (secret.domain,
                  secret.access,
                  secret.user_id,
                  secret.password,
                  json.dumps(secret.info),
                  timestamp)
        query = f"""INSERT INTO [{self.table_name}] 
                    VALUES (?1,?2,?3,?4,?5,?6) 
                    ON CONFLICT (domain, access)
                    DO UPDATE
                    SET uid = ?3,
                        pwd = ?4,
                        info = ?5,
                        timestamp = ?6;                    
                """
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,record)
            connection.commit()
        finally:
            cursor.close()
            connection.close


    def get_record(self, secret:Secret):
        "retrieves an encrypted record keyed by domain and access, as a dictionary"
        query = f"""SELECT * FROM [{self.table_name}]
                    WHERE domain = ?1
                    AND   access = ?2;
                """
        in_record = (secret.domain, secret.access)
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,in_record)
            out_record = cursor.fetchone()
            if out_record is None:
                return None
            return Secret(domain=out_record[0],
                          access = out_record[1],
                          user_id = out_record[2],
                          password = out_record[3],
                          info = json.loads(out_record[4]),
                          timestamp = out_record[5])
        finally:
            cursor.close()
            connection.close


    def update_record_single_info(self, 
                                  secret:Secret, 
                                  mem_pwd:str, 
                                  salt:str, 
                                  timestamp:str=datetime.now().isoformat()) -> None:
        "update a record in the secrets table, with a maximum of one extra info stored in the Secret record"
        
        query = f"UPDATE [{self.table_name}]"
        params = { "domain" : secret.domain, "access" : secret.access, "timestamp" : timestamp}    
        condition = " WHERE domain=:domain AND access=:access;"
        change = False
        update = "SET timestamp = :timestamp"
        if secret.user_id is not None:
            change = True
            params["uid"] = encrypt(secret.user_id, mem_pwd, salt)
            update += ", uid = :uid"
        if secret.password is not None:
            change = True
            params["pwd"] = encrypt(secret.password, mem_pwd, salt)
            update += ", pwd = :pwd"
        if secret.info_key is not None and secret.info_value is not None:
            change = True
            old_secret = self.get_record(secret)
            if old_secret is not None:
                old_secret.info[secret.info_key] = encrypt(secret.info_value, mem_pwd, salt)
                params["info"] = json.dumps(old_secret.info)
                update += ", info = :info"
                
        if not change: #nothing to do
            return
        query += update + condition
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,params)
            connection.commit()
        finally:
            cursor.close()
            connection.close        


    def update_record_info_dictionary(self, 
                                      secret:Secret, 
                                      timestamp:str=datetime.now().isoformat()) -> None:
        "update the info dictionary for a record in the secrets table (pre-encrypted)"
        einfo = secret.encrypted_info
        if einfo is None:
            return
        
        query = f"""UPDATE [{self.table_name}]
                    SET info =:info, timestamp=:timestamp
                    WHERE domain=:domain AND access=:access;
                """
        
        params = { "domain" : secret.domain, 
                   "access" : secret.access,
                   "info"   : json.dumps(einfo), 
                   "timestamp" : timestamp}
        
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,params)
            connection.commit()
        finally:
            cursor.close()
            connection.close        


    def delete_record(self, secret:Secret)->None:
        "delete a single record based on domain and access keys"
        query = f"""DELETE from [{self.table_name}]
                    WHERE domain = ?1
                    AND access = ?2;"""
        in_record = (secret.domain, secret.access)
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            cursor.execute(query,in_record)
            connection.commit()
        finally:
            cursor.close()
            connection.close

    def query_record(self, secret:Secret)-> list:
        "return a list of secrets matching the domain key passed"
        where_clause = ""
        binding = False
        if secret.domain is not None:
            where_clause = "WHERE domain = ?"
            binding = True #there is a parameter to bind        
        query = f"""SELECT domain, access from [{self.table_name}] {where_clause};"""
        
        ret = list()
        try:
            connection = sqlite3.connect(self.db_path)
            cursor = connection.cursor()
            if binding:
                cursor.execute(query,(secret.domain,))
            else:
                cursor.execute(query)
            records = cursor.fetchall()
            for r in records:
                ret.append(Secret(domain=r[0], access=r[1]))
        finally:
            cursor.close()
            connection.close        
        
        return ret




'''
Created on 1 Jan 2020

@author: codimoc
'''

import argparse
import contextlib
import json
import shlex
import sys

import readline
from secretwallet.constants import parameters, Secret
from secretwallet.main.configuration import list_configuration, get_configuration, set_configuration_data
from secretwallet.session.client import get_session_password, set_session_password, stop_service, is_connected
from secretwallet.session.service import start_my_session
from secretwallet.utils.cryptutils import encrypt_key
from secretwallet.utils.dbutils import has_secret, get_secret, insert_secret, insert_encrypted_secret, list_secrets, \
                                       update_secret, update_secret_info_dictionary, delete_secret, delete_secrets, rename_secret, \
                                       reconf_memorable, reconf_salt_key, query_secrets_by_field, query_secrets_by_pattern, \
                                       get_all_secrets, has_table, create_table
from secretwallet.utils.logging import get_logger

import pkg_resources as pkg
import secretwallet.utils.ioutils as iou
import secretwallet.utils.password_manager as pm


logger = get_logger(__name__, parameters.get_log_level())
parameters.register_logger(__name__, logger)

usage_bash = '''secret_wallet <command> [<args>]

The list of secretwallet commands are:
   set             Insert a new secret
   get             Retrieves a secret
   delete          Remove a secret
   rename          rename a secret
   list            list all secrets in a given domain
   conf            manage the configuration file
   query           query secrets based on a condition
   qget            query based on pattern, and retrieval of chosen secret
   reconf          change an existing configuration
   session         (testing) start a session to store the memorable password between consecutive calls
   client          (testing) retrieves the memorable password from the running session
   help            print the main help page
   version         the version of this package
   shell           starts a secret_wallet sheel for interctive queries
   dump            dump all secrets to a file in text format
   save            save all secrets to a json file for backup and safe keeping (to be re-loaded)
   load            reload all secrets from a json file (backup and restore)
   ....

For individual help type:
secretwallet <command> -h
'''

usage_shell = '''<command> [<args>]

The list of secretwallet commands are:
   set             Insert a new secret
   get             Retrieves a secret
   delete          Remove a secret
   rename          rename a secret
   list            list all secrets in a given domain
   conf            manage the configuration file
   query           query secrets based on a condition
   qget            query based on pattern, and retrieval of chosen secret
   reconf          change an existing configuration
   help            print the main help page
   version         the version of this package
   quit            terminate the interactive shell
   dump            dump all secrets to a file in text format
   save            save all secrets to a json file for backup and safe keeping (to be re-loaded)
   load            reload all secrets from a json file (backup and restore)
   ....

For individual help type:
<command> -h
'''

class Parser(object):

    def __init__(self):
        parser = argparse.ArgumentParser(
            description='The Secrets manager',
            usage=usage_bash)
        parser.add_argument('command',
                            action='store',
                            choices=['set','get','delete', 'rename', 'list', 'conf',
                                     'query','qget','reconf','help','session','client',
                                     'version', 'shell', 'dump', 'save','load'],
                            help='Command to run')
        self._parser = parser
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            iou.my_output('Unrecognized command')
            parser.print_help()
            exit(1)
        
        #create table if not there for the first time
        create_table(silent=True) #if table already there nothing happens
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()


    def set(self):
        """
            Add or change a secret in the wallet. This could be an entire new secret, with all the information passed inline
            or an update of an existing secret. What is set with this command determines the content of a secret, identified
            by the domain, access pair. Key values pairs, as defined by the -ik and -iv options, can be added incrementally
            by multiple calls to the set command.
        """
        parser = argparse.ArgumentParser(
            description=self.set.__doc__,
            prog='secret_wallet set')
        #required arguments
        parser.add_argument('-d',
                            dest='domain',
                            required=True,
                            help='The domain (category) of the secret')
        parser.add_argument('-a',
                        dest='access',
                        required=True,
                        help='The sub=domain (sub-category or access) of the secret')
        #optional arguments
        parser.add_argument('-u',
                            '--uid',
                            help='The login id for a given access')
        parser.add_argument('-p',
                            '--pwd',
                            help='The password for a given access')
        parser.add_argument('-ik',
                            '--info_key',
                            help='The key in an information map')
        parser.add_argument('-iv',
                            '--info_value',
                            help='The value in an information map')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Running set for domain %s and access %s' %(args.domain,args.access))
        if args.info_key is None or args.info_value is None:
            info = None
        else:
            info = {args.info_key :args.info_value}
        try:
            memorable, need_session = pm.get_memorable_password(True)
            if not has_secret(args.domain, args.access):
                insert_secret(args.domain, args.access, args.uid, args.pwd, info , memorable)
            else:
                update_secret(args.domain, args.access, args.uid, args.pwd, args.info_key, args.info_value, memorable)
            if need_session:
                start_my_session(memorable, parameters.get_session_lifetime(), parameters.get_session_timeout())
        except Exception as e:
            iou.my_output(repr(e))

    def get(self):
        """
            Retrieves the information stored inside a secret, as identified by the domain, access pair. These two fields need
            to be passed by using the -d and -a options
        """
        parser = argparse.ArgumentParser(
            description=self.get.__doc__,
            prog='secret_wallet get')
        #required arguments
        parser.add_argument('-d',
                            dest='domain',
                            required=True,
                            help='The domain (category) of the secret')
        parser.add_argument('-a',
                        dest='access',
                        required=True,
                        help='The sub=domain (sub-category or access) of the secret')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Running get with arguments %s' % args)
        try:
            memorable, need_session = pm.get_memorable_password(False)
            iou.display_secret(get_secret(args.domain, args.access, memorable))
            if need_session:
                start_my_session(memorable, parameters.get_session_lifetime(), parameters.get_session_timeout())
        except Exception as e:
            iou.my_output(repr(e))

    def delete(self):
        """
            Deletes an existing secret, as identified by the domain, access pair. These two fields can
            be passed by using the -d and -a options. If only the domain is given, all secrets for that domain
            are deleted. When the -ik option is given with a key name, the corresponding entry
            in the info dictionary is removed, only if both domain and access are given and they identify an existing secret
        """
        parser = argparse.ArgumentParser(
                              description=self.delete.__doc__,
                              prog='secret_wallet delete')
        #required arguments
        parser.add_argument('-d',
                            dest='domain',
                            required=True,
                            help='The domain (category) of the secret')
        parser.add_argument('-a',
                            dest ='access',
                            help='The sub=domain (sub-category or access) of the secret')
        parser.add_argument('-ik',
                            '--info_key',
                            help='The key in an information map')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Running delete with arguments %s' % args)
        try:
            if args.domain is not None and args.access is not None and args.info_key is not None:
                iou.confirm_delete_key(args.domain, args.access, args.info_key)
                sec = get_secret(args.domain, args.access, None, None, False) #no decryption
                info = sec.info
                del info[args.info_key]
                #Here the dictionary passed is already encrypted
                update_secret_info_dictionary(args.domain, args.access, info)
            elif args.domain is not None and args.access is not None:
                secret = Secret(domain=args.domain, access=args.access)
                iou.confirm_delete([secret])
                delete_secret(args.domain, args.access)
            else:
                secrets = list_secrets(args.domain)
                iou.confirm_delete(secrets)
                delete_secrets(secrets)
        except Exception as e:
            iou.my_output(repr(e))

    def rename(self):
        """
           Renames a secret, as identified by the domain, access pair. A new domain name can be passed with the -nd option and a new access
           name can be passed with the -na option. Both domain and access can be changed at the same time or on their own.
        """
        parser = argparse.ArgumentParser(
            description=self.rename.__doc__,
            prog='secret_wallet rename')
        #required arguments
        parser.add_argument('-d',
                            dest='domain',
                            required=True,
                            help='The domain (category) of the secret')
        parser.add_argument('-a',
                            dest ='access',
                            required=True,
                            help='The sub=domain (sub-category or access) of the secret')
        parser.add_argument('-nd',
                            dest='new_domain',
                            default = None,
                            help='The new domain')
        parser.add_argument('-na',
                            dest ='new_access',
                            default = None,
                            help='The new asset')
        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Running rename with arguments %s' % args)
        try:
            if args.new_domain is None and args.new_access is None:
                iou.my_output("No new keys have been passed", True)
            elif args.new_domain == args.domain and args.new_access == args.access:
                iou.my_output("Both new values are the same as the originals: nothing to do", True)
            elif not has_secret(args.domain, args.access):
                iou.my_output("Could not find the secret to rename", True)
            else:
                if args.new_domain is None:
                    args.new_domain = args.domain
                if args.new_access is None:
                    args.new_access = args.access
                secret = Secret(domain=args.domain, access=args.access)
                iou.confirm_rename([secret])
                rename_secret(args.domain, args.access, args.new_domain, args.new_access)
        except Exception as e:
            iou.my_output(repr(e))

    def list(self):
        """
           List a set of secrets. With no option passed, all secrets are returned. Alternatively it is possible to filter secrets by
           passing a domain name: all secrets for that domain will be returned.
        """
        parser = argparse.ArgumentParser(
            description=self.list.__doc__,
            prog='secret_wallet list')
        #optional arguments
        parser.add_argument('-d',
                            '--domain',
                            help='The domain (category) of the secrets. If not given all secrets are returned')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Running list with arguments %s' % args)
        try:
            secrets = list_secrets(args.domain)
            iou.display_list("List of secrets", secrets)
        except Exception as e:
            iou.my_output(repr(e))

    def query(self):
        """
           Searches for secrets containig a given subtext in either their domain or access names, or both. By using the explicit -d and -a
           options, it is possible to limit the search to domain names or access names only. Alternatively it is possible to pass a subtext
           without any specification in front (i.e. without -d or -a) and the search of that pattern will include both domain and access names.
        """
        parser = argparse.ArgumentParser(
            description=self.query.__doc__,
            prog='secret_wallet query')
        #optional arguments
        parser.add_argument('-d',
                            '--domain',
                            default=None,
                            help='A substring to query the domains. Only secrets with this substring in their domain are returned')
        parser.add_argument('-a',
                            '--access',
                            default=None,
                            help='A substring to query the access. Only secrets with this substring in their access are returned')

        parser.add_argument('pattern',
                            nargs='?',
                            default=None,
                            help='A pattern that is searched both in the domain and the access field')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Query secrets with arguments %s' % args)
        try:
            if (args.pattern is not None):
                secrets = query_secrets_by_pattern(args.pattern)
            else:
                secrets = query_secrets_by_field(args.domain, args.access)
            iou.display_list("List of secrets", secrets)
        except Exception as e:
            iou.my_output(repr(e))

    def qget(self):
        """
           Searches for secrets containig a given subtext in their domain or access names. Once a list of secrets
           that match the given pattern is found, thre secrets are tagged with a progressive number and the user can
           select the one to retrieve and display.
        """
        parser = argparse.ArgumentParser(
            description=self.qget.__doc__,
            prog='secret_wallet qget')

        parser.add_argument('pattern',
                            default=None,
                            help='A pattern that is searched both in the domain and the access field')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Query secrets with arguments %s' % args)
        try:
            if (args.pattern is not None):
                secrets = query_secrets_by_pattern(args.pattern)
            sec = iou.get_secret_by_idx("List of secrets", secrets)
            if sec is None:
                return
            try:
                memorable, need_session = pm.get_memorable_password(False)
                iou.display_secret(get_secret(sec.domain, sec.access, memorable))
                if need_session:
                    start_my_session(memorable, parameters.get_session_lifetime(), parameters.get_session_timeout())
            except Exception as e:
                iou.my_output(repr(e))

        except Exception as e:
            iou.my_output(repr(e))

    def conf(self):
        """
           Configures some parameters for this application. It is possible to list all parameters with the -l option,
           or to configure the timeout and lifetime (in seconds) or the log level.
           The timeout is the amount of time in seconds along which the memorable password is remembered without been re-asked.
           The lifetime determines the lifetime of the background process that manages the temporary storage of
           the memorable password. The value of the lifetime parameter should be bigger than the password timeout.
           The logging level is one of debug, info, warning, critical, error or fatal. The srorage_type determines the
           type of storage used: aws_dynamo for remote storage on AWS or local_sqlite for SQLite based local storage.
        """
        parser = argparse.ArgumentParser(
            description=self.conf.__doc__,
            prog='secret_wallet conf')
        #optional arguments
        parser.add_argument('-l',
                            '--list',
                            action = 'store_true',
                            default = False,
                            help='List the existing configuration parameters')
        parser.add_argument('-to',
                            '--timeout',
                            dest = 'timeout',
                            type = int,
                            default = -1,
                            help='Session time-out in seconds')
        parser.add_argument('-lf',
                            '--lifetime',
                            dest = 'lifetime',
                            type = int,
                            default = -1,
                            help='Session lifetime in seconds')
        parser.add_argument('-ll',
                            '--loglevel',
                            dest = 'loglevel',
                            help='Logging level. One of debug, info, warning, critical, error or fatal')
        parser.add_argument('-st',
                            '--storage',
                            dest = 'storage',
                            help='Storage type: aws_dynamo for remote storage on AWS, local_sqlite for local storage')        

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Running conf with arguments %s' % args)
        try:
            if (args.list):
                list_configuration()
            else:
                conf = get_configuration()
                if args.timeout is not None and args.timeout >=0:
                    conf['session_timeout']=args.timeout
                if args.lifetime is not None and args.lifetime >=0:
                    conf['session_lifetime']=args.lifetime
                if args.loglevel is not None:
                    if args.loglevel.lower() in ['debug', 'info', 'warning', 'critical', 'error','fatal']:
                        conf['log_level'] = args.loglevel.lower()
                    else:
                        iou.my_output(f"The passed log level {args.loglevel.lower()} is not valid")
                if args.storage is not None:
                    if args.storage.lower() in ['aws_dynamo', 'local_sqlite']:
                        conf['storage_type'] = args.storage.lower()
                    else:
                        iou.my_output(f"The passed storage_type {args.storage.lower()} is not valid")                        
                set_configuration_data(conf)
        except Exception as e:
            iou.my_output(repr(e))

    def reconf(self):
        """
           Reconfigures either the memorable or the device password. All secrets will be re-encryted with the changed password.
           It is not possible to change both passwords at the same time. Depending on the size of the wallet, this operation
           might take some time. A backup of the old table is also performed.
        """
        parser = argparse.ArgumentParser(
            description=self.reconf.__doc__,
            prog='secret_wallet reconf')
        #optional arguments
        parser.add_argument('-m',
                            '--memorable',
                            action = 'store_true',
                            default = False,
                            help='Reconfigure secrets because of a change of memorable password')
        parser.add_argument('-d',
                            '--device',
                            action = 'store_true',
                            default = False,
                            help='Reconfigure secrets because of a change of device password')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        try:
            if args.memorable and args.device:
                print("You can't reconfigure both memorable and device password at the same time")
            elif args.memorable:
                iou.my_output("You are reconfiguring the memorable password")
                if is_connected():
                    stop_service()
                iou.display_reconfiguration_warning()

                iou.my_output('***Enter the existing memorable password***')
                old_memorable, _ = pm.get_memorable_password(False)
                iou.my_output('***Set the new memorable password***')
                new_memorable, _ = pm.get_memorable_password(True)
                reconf_memorable(list_secrets(None), old_memorable, new_memorable, True)
            elif args.device:
                iou.my_output("You are reconfiguring the device password")
                if is_connected():
                    stop_service()
                iou.display_reconfiguration_warning()

                iou.my_output('***Enter the existing memorable password***')
                old_memorable, _ = pm.get_memorable_password(False)
                iou.my_output('***Set the new device password***')
                new_device, _ = pm.get_memorable_password(True)
                reconf_salt_key(list_secrets(None), old_memorable, new_device, True)

                #now pass it to the configuration file
                ekey = encrypt_key(new_device)
                cdata = get_configuration()
                cdata['key'] = ekey
                set_configuration_data(cdata)
        except Exception as e:
            iou.my_output(repr(e))

    def help(self):
        self._parser.print_help()

    def version(self):
        print("secret-wallet-codimoc version %s"%pkg.get_distribution('secret-wallet-codimoc').version)

    def shell(self):
        parser = argparse.ArgumentParser(
            description='Interactive secret_wallet shell',
            prog='secret_wallet shell')
        parser.add_argument('command',
                            action='store',
                            choices=['set','get','delete', 'rename', 'list', 'conf',
                                     'query','qget','reconf','help','session','quit',
                                     'version', 'dump', 'save'],
                            help='Command to run inside the shell')
        iou.my_output('Starting a secret_wallet interactive shell. Type quit to quit, help for help')
        parameters.set_in_shell(True)
        #using readline for history of command line and other
        readline.parse_and_bind('tab: complete')
        readline.parse_and_bind('set editing-mode vi')
        while True: #this is the shell main loop
            cmd = iou.my_input(':> ')
            if cmd.lower().startswith('quit'):
                parameters.set_in_shell(False)
                break
            if cmd.lower().startswith('help'):
                iou.my_output(usage_shell, with_logging=False)
                continue
            tokens = shlex.split(cmd) #keep quoted test together
            try:
                args = parser.parse_args(tokens[:1])
            except SystemExit as e: #don't break the shell
                parameters.set_in_shell(False)
                iou.my_output("Wrong Input command!!")
                iou.my_output(usage_shell, with_logging=False)
                continue
            if not hasattr(self, args.command):
                iou.my_output('Unrecognized command')
                iou.my_output(usage_shell, with_logging=False)
                continue
            # use dispatch pattern to invoke method with same name
            try:
                sys.argv=['secret_wallet'] + tokens  #append a first argument just for padding (could be anything)
                parameters.set_in_shell(True)
                getattr(self, args.command)()
                parameters.set_in_shell(False)
            except Exception as e: #don't break the shell
                parameters.set_in_shell(False)
                iou.my_output(repr(e))
                continue

    def session(self):
        """
           Starts a background session for keeping track of the memorable password for a short while. This is only for testing
           since this process is started automatically when needed by the secret_wallet. The lifetime parameter sets the lifetime of
           the session in seconds, the timeout the time in second for which the memorable password is kept, and the value is what
           has to be remembered
        """
        parser = argparse.ArgumentParser(
            description=self.session.__doc__,
            prog='secret_wallet session')
        #optional arguments
        parser.add_argument('-l',
                            dest = 'lifetime',
                            type = int,
                            help='The lifetime in seconds of the session',
                            default = parameters.get_session_lifetime())
        parser.add_argument('-t',
                            dest = 'timeout',
                            type = int,
                            help='The timeout in seconds for the session value',
                            default = parameters.get_session_timeout())
        parser.add_argument('-v',
                            dest = 'value',
                            help='The value to store in the session',
                            default='not set')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Starting a secret wallet session with parameters %s'%args)
        try:
            start_my_session(args.value, args.lifetime, args.timeout)
        except Exception as e:
            iou.my_output(repr(e))


    def client(self):
        """ Client command to invoke the background session. This is for testing only. The action allows to get the session value,
            set the value, stop the background session and test if it is running
        """
        parser = argparse.ArgumentParser(
            description=self.client.__doc__,
            prog='secretwallet client')
        #optional arguments
        parser.add_argument('-a',
                            dest = 'action',
                            choices=['get','set','stop','test'],
                            help='The client action',
                            default = 'get')
        parser.add_argument('-v',
                            dest = 'value',
                            help='The value to store in the session',
                            default='not set')

        args = iou.my_parse(parser,sys.argv[2:])
        if args is None:
            return

        iou.my_output('Starting a secret wallet client with parameters %s'%args)
        try:
            if args.action == 'get':
                iou.my_output((get_session_password()[0],'***'))
            elif args.action == 'set':
                set_session_password(args.value)
            elif args.action == 'stop':
                stop_service()
            elif args.action == 'test':
                if is_connected():
                    iou.my_output('connected')
                else:
                    iou.my_output('not connected')
        except Exception as e:
            iou.my_output(repr(e))

    def dump(self):
        " Dumps all secrets to a flat file , designated by the option -f, or the console when not specified."
        parser = argparse.ArgumentParser(
            description=self.dump.__doc__,
            prog='secretwallet dump')
        #optional arguments
        parser.add_argument('-f',
                            dest='file',
                            required=False,
                            help='The output file. If it is not specified it dumps to the console (default system out stream)')

        args = iou.my_parse(parser,sys.argv[2:])

        iou.my_output('Starting a secret wallet dump with parameters %s'%args)
        try:
            memorable, need_session = pm.get_memorable_password(False)

            secrets = get_all_secrets(memorable,False) #False -> return as Secret
            if (args.file is not None):
                with open(args.file,'a') as f:
                    with contextlib.redirect_stdout(f):
                        iou.display_all_secrets(secrets) #send to file
            else:
                iou.display_all_secrets(secrets) #send to the console

            #here start the session (at the end so that we can daemonize)
            if need_session:
                start_my_session(memorable, parameters.get_session_lifetime(), parameters.get_session_timeout())
        except Exception as e:
            iou.my_output(repr(e))

    def save(self):
        """ Save all secrets to a json file , designated by the option -f, or the console when not specified.
        This file can be used for safe-keeping and reload, via the load function
        """
        parser = argparse.ArgumentParser(
            description=self.save.__doc__,
            prog='secretwallet save')
        #optional arguments
        parser.add_argument('-f',
                            dest='file',
                            required=False,
                            help='The output file. If it is not specified it outputs the json structure to the console (default system out stream)')
        parser.add_argument('-d',
                        dest='needs_decrypt',
                        required=False,
                        default=False,
                        action='store_true',
                        help='A flag. If set, all secret are stored or displayed after being decrypted. It defaults to encrypted if not passed')

        args = iou.my_parse(parser,sys.argv[2:])

        iou.my_output('Starting a secret wallet save with parameters %s'%args)
        try:
            if args.needs_decrypt:
                memorable, need_session = pm.get_memorable_password(False)
    
                secrets = get_all_secrets(memorable, True) # True -> return as dictionary
                if (args.file is not None):
                    with open(args.file,'w') as f:
                        with contextlib.redirect_stdout(f):
                            json.dump(secrets, f) #send to file
                else:
                    iou.my_output(json.dumps(secrets), with_logging=False)
    
                #here start the session (at the end so that we can daemonize)
                if need_session:
                    start_my_session(memorable, parameters.get_session_lifetime(), parameters.get_session_timeout())
            else: # encrypted
                secrets = get_all_secrets(as_dictionary=True, needs_decrypt=False)
                if (args.file is not None):
                    with open(args.file,'w') as f:
                        with contextlib.redirect_stdout(f):
                            json.dump(secrets, f) #send to file
                else:
                    iou.my_output(json.dumps(secrets), with_logging=False)
    
        except Exception as e:
            iou.my_output(repr(e))

    def load(self):
        """Reload secrets, previously stored in a json file after a backup operation (save action).
        It requires a json file as input, as specified with the -f option."""
        parser = argparse.ArgumentParser(
            description=self.load.__doc__,
            prog='secretwallet load')
        #optional arguments
        parser.add_argument('-f',
                            dest='file',
                            required=True,
                            help='The input file in json format, containing the backed-up secrets.')
        parser.add_argument('-d',
                            dest='needs_decrypt',
                            required=False,
                            default=False,
                            action='store_true',
                            help='A flag. If set, all secret are loaded from decrypted values.  It defaults to encrypted if not passed')        

        args = iou.my_parse(parser,sys.argv[2:])
        iou.my_output('Starting a secret wallet load with parameters %s'%args)
        try:
            if args.needs_decrypt:
                memorable, need_session = pm.get_memorable_password(True)
    
                with open(args.file,'r') as f:
                    secrets = json.load(f)
                    for secret in secrets:
                        domain = secret["domain"]
                        access = secret["access"]
                        if not has_secret(domain, access):
                            iou.my_output(f"inserting secret (domain:{domain}, access:{access})")
                            insert_secret(domain, access, secret["uid"], secret["pwd"], secret["info"] , memorable, timestamp=secret["timestamp"])
    
                #here start the session (at the end so that we can daemonize)
                if need_session:
                    start_my_session(memorable, parameters.get_session_lifetime(), parameters.get_session_timeout())
            else: #encrypted
                with open(args.file,'r') as f:
                    secrets = json.load(f)
                    for secret in secrets:
                        domain = secret["domain"]
                        access = secret["access"]
                        if not has_secret(domain, access):
                            iou.my_output(f"inserting secret (domain:{domain}, access:{access})")
                            insert_encrypted_secret(domain, access, secret["uid"], secret["pwd"], secret["info"],timestamp=secret["timestamp"])
        except Exception as e:
            iou.my_output(repr(e))



'''
Created on 24 Dec 2019

@author: codimoc
'''

from datetime import datetime
import sys

from secretwallet.constants import parameters, Secret, DB_AWS_DYNAMO, DB_LOCAL_SQLITE
from secretwallet.storage.aws_dynamo import AWSDynamoTable
from secretwallet.storage.local_sqlite import LocalSqLiteTable
from secretwallet.utils.cryptutils import encrypt_key, decrypt, decrypt_info
from secretwallet.utils.logging import get_logger

from secretwallet.storage.table import Table
import secretwallet.utils.ioutils as iou


logger = get_logger(__name__, parameters.get_log_level())
parameters.register_logger(__name__, logger)

SEPARATOR="#-#"

def secret_to_dictionary(secret:Secret, mem_pwd = None, salt = None, needs_decrypt:bool = True):
    ret = {'domain'    : secret.domain,
           'access'    : secret.access,
           'timestamp' : secret.timestamp}
    if needs_decrypt:
        ret['uid']  = decrypt(secret.user_id, mem_pwd, salt)
        ret['pwd']  = decrypt(secret.password, mem_pwd, salt)
        ret['info'] = decrypt_info(secret.info, mem_pwd, salt)
    else:
        ret['uid']  = secret.user_id
        ret['pwd']  = secret.password
        ret['info'] = secret.info
    return ret

def decrypt_secret(secret:Secret, mem_pwd:str, salt:str)->Secret:
    return Secret(domain = secret.domain,
                  access = secret.access,
                  user_id = decrypt(secret.user_id, mem_pwd, salt),
                  password = decrypt(secret.password, mem_pwd, salt),
                  info = decrypt_info(secret.info, mem_pwd, salt),
                  timestamp = secret.timestamp
                  )

def _get_table()->Table:
    if parameters.get_storage_type() == DB_AWS_DYNAMO: 
        return AWSDynamoTable(parameters.get_table_name(), parameters.get_profile_name())
    elif parameters.get_storage_type() == DB_LOCAL_SQLITE:
        return LocalSqLiteTable(parameters.get_table_name())

def _backup_table(backup_name:str)->object:
    return _get_table().backup_table(backup_name)

def _cleanup_table_backups(backup_name:str)->None:
    _get_table().cleanup_table_backups(backup_name)

def _drop_table():
    _get_table().drop_table()

def has_table(table_name:str)->bool:
    "Checks if the table exists"
    try:
        return _get_table().has_table(table_name)
    except Exception as e:
        iou.my_output(e)
        logger.error(e)
        sys.exit(1)


def create_table(table_name=parameters.get_table_name(), silent=False):
    "Creates a table if it does not exist"
    try:
        _get_table().create_table(table_name)
    except Exception as e:
        logger.error(e)
    if has_table(table_name) and not silent:
        logger.info(f"Table {table_name} has been created")
        iou.my_output(f"Table {table_name} has been created")


def insert_secret(domain, access, uid, pwd, info, mem_pwd, salt=None, timestamp = None):
    """Insert a secret access record in the cloud DB
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    uid        the user id for that access
    pwd        the password for that access
    info       a map of informations (not encryted)
    mem_pwd    memorable password to encrypt the secret
    salt       a string representation of the salt (optional)
    timestamp  the secret timestamp. If not passed it is calculated
    """
    if salt is None:
        salt = parameters.get_salt_key()
    if timestamp is None:
        timestamp = datetime.now().isoformat()
    if uid is None:
        uid = ""
    if pwd is None:
        pwd = ""
    if info is None:
        info = {}
    secret = Secret(domain = domain,
                    access = access,
                    user_id = uid,
                    password = pwd,
                    info = info)
    _get_table().insert_record(secret, mem_pwd, salt, timestamp)
    
def insert_encrypted_secret(domain, access, uid, pwd, info, timestamp = None):
    """Insert a secret access record in the cloud DB
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    uid        the encrypted user id for that access
    pwd        the encrypted password for that access
    info       a map of informations (encryted)
    timestamp  the secret timestamp. If not passed it is calculated
    """
    if timestamp is None:
        timestamp = datetime.now().isoformat()
    if uid is None:
        uid = ""
    if pwd is None:
        pwd = ""
    if info is None:
        info = {}
    secret = Secret(domain = domain,
                    access = access,
                    user_id = uid,
                    password = pwd,
                    info = info)
    _get_table().insert_encrypted_record(secret,timestamp)    

def update_secret(domain, access, uid, pwd, info_key, info_value, mem_pwd, salt=None):
    """Update a secret access record in the cloud DB
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    uid        the user id for that access
    pwd        the password for that access
    info_key   the key for an extra info
    info_value the value for an extra info
    mem_pwd    memorable password to encrypt the secret
    salt       a string representation of the salt (optional)
    """
    if salt is None:
        salt = parameters.get_salt_key()
    timestamp = datetime.now().isoformat()
    secret = Secret(domain = domain,
                    access = access,
                    user_id = uid,
                    password = pwd,
                    info_key = info_key,
                    info_value = info_value)    
    try:
        _get_table().update_record_single_info(secret, mem_pwd, salt, timestamp)
    except:
        pass #the condition failed but there should be no side effect

def update_secret_info_dictionary(domain, access, enc_info):
    """Update the info dictionary of a secret
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    enc_info   the new info dictionary (encrypted) to replace the old one
    """

    timestamp = datetime.now().isoformat()
    try:
        secret = Secret(domain = domain,
                        access = access,
                        encrypted_info = enc_info)
        _get_table().update_record_info_dictionary(secret, timestamp)
    except:
        pass #the condition failed but there should be no side effect


def rename_secret(domain, access, new_domain, new_access):
    """Rename the domain and access of a secret
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    new_domain the new value for the domain
    new_access the new value for the access
    """
    table = _get_table()
    filter_secret = Secret(domain = domain, access = access)
    old_secret = table.get_record(filter_secret)
    if old_secret is not None:
        try:
            new_secret = Secret(domain = new_domain,
                                access = new_access,
                                user_id = old_secret.user_id,
                                password = old_secret.password,
                                info = old_secret.info)
            table.insert_encrypted_record(new_secret, datetime.now().isoformat())
            table.delete_record(filter_secret)
        except Exception as e:
            iou.my_output(e)
    else:
        iou.my_output(f"Could not find secret ({domain},{access})")

def has_secret(domain, access):
    """Checks the existence of a secret
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    """
    filter_secret = Secret(domain = domain, access = access)
    secret = _get_table().get_record(filter_secret)
    return secret is not None

def delete_secret(domain, access):
    """Delete a secret by primary key.
    input:
    domain     the domain, i.e. logical context, of the secret
    access     the secret sub-domain or access specification
    """
    filter_secret = Secret(domain = domain, access = access)
    _get_table().delete_record(filter_secret)

def delete_secrets(secrets):
    """Deletes all secrets passed as list of (domain, access) pairs
    input:
    secrets    a list of secrets, as Secret objects
    """
    for s in secrets:
        delete_secret(s.domain, s.access)

def get_secret(domain, access, mem_pwd=None, salt=None, needs_decrypt=True):
    """Retrieves a secret by primary key
    input:
    domain          the domain, i.e. logical context, of the secret
    access          the secret sub-domain or access specification
    mem_pwd         memorable password to encrypt the secret
    salt            a string representation of the salt (optional)
    need_decrypt    a flag to indicate if decryption is required (default True)
    output:
    returns the decrypted secret as a dictionary
    """
    if salt is None:
        salt = parameters.get_salt_key()

    filter_secret = Secret(domain = domain, access = access)
    secret = _get_table().get_record(filter_secret)
    if needs_decrypt:
        return decrypt_secret(secret, mem_pwd, salt)
    else:
        return secret


def list_secrets(domain):
    """List all secrets by domain
    input:
    domain    the domain of the secrets. If null all records are returned
    output:
    a list of Secret objects
    """
    filter_secret = Secret(domain=domain)
    return _get_table().query_record(filter_secret)

def get_all_secrets(mem_pwd=None, as_dictionary=True, needs_decrypt = True):
    """Get all secrets and return them as a list of dictionaries or secret objects
    input:
    mem_pwd        the memorable password
    as_dictionary  a flag to select the type ot representation: dictionary or Secret
    needs_decrypt  a flag indicatinf that the returned secrets should be decrypted
    output:
    a list of secrets as dictionaries or secret objects depending on the flag passed
    """
    secrets = []
    for s in list_secrets(None):
        if as_dictionary:
            secret = get_secret(s.domain, s.access, mem_pwd, None, needs_decrypt) #decrypted
            secrets.append(secret_to_dictionary(secret, needs_decrypt=False))
        else:
            secrets.append(get_secret(s.domain, s.access, mem_pwd, None, needs_decrypt)) #decrypted
    return secrets

def query_secrets_by_field(domain_sub, access_sub):
    """Query all secrets with domain and access containing the domain or access substrings
    input:
    domain_sub  the substring to be looked for in the domain keys
    access_sub  the substring to be looked for in the access keys
    output:
    a list of Secret objects
    """
    secrets = list_secrets(None)
    filter_secrets = lambda s:(domain_sub is None or domain_sub.lower() in s.domain.lower())\
                               and\
                               (access_sub is None or access_sub.lower() in s.access.lower())
    return [s for s in secrets if filter_secrets(s)]

def query_secrets_by_pattern(pattern):
    """Query all secrets with domain or access containing the pattern substrings
    input:
    pattern  the substring to be looked for in the domain or access field
    output:
    a list of Secret objects
    """
    secrets = list_secrets(None)
    lpt = pattern.lower()
    filter_secrets = lambda s: pattern is None or lpt in s.domain.lower() or lpt in s.access.lower()

    return [s for s in secrets if filter_secrets(s)]

def reconf_memorable(secrets, old_mem, new_mem, backup=False):
    """Reconfigure all secrets changing the memorable password
    input:
    secrets    a list of Secret objects
    old_mem    old memorable password
    new_mem    new memorable password
    backup     a boolean flag to request a full baclup of the table
    output:
    the BackupArn of the table
    """
    arn = None
    ns = len(secrets)
    i = 0
    if backup:
        arn = _get_table().backup_table("backup")
    for s in secrets:
        i+=1
        domain = s.domain
        access = s.access
        try:
            message = f"[{i}/{ns}] - Reconfiguring the secret ({domain},{access})"
            logger.info(message)
            iou.my_output(message)
            secret = get_secret(domain, access, old_mem)
            insert_secret("I", f"{domain}{SEPARATOR}{access}", secret.user_id,  secret.password,  secret.info, new_mem)
            rename_secret(domain, access, "D", f"{domain}{SEPARATOR}{access}")
            rename_secret("I", f"{domain}{SEPARATOR}{access}", domain, access)
            delete_secret("D", f"{domain}{SEPARATOR}{access}")
        except Exception as e:
            logger.error(e)
            message = f"Could not reconfigure ({domain},{access})"
            iou.my_output(message)
            logger.error(message)
    return arn

def reconf_salt_key(secrets, old_mem, new_device_pwd, backup=False):
    """Reconfigure all secrets changing the memorable password
    input:
    secrets         a list of Secret objects
    old_mem         old memorable password
    new_device_pwd  the new device password
    backup          a boolean flag to request a full baclup of the table
    output:
    the BackupArn of the table
    """
    ekey = encrypt_key(new_device_pwd)
    ns = len(secrets)
    i = 0
    arn = None
    if backup:
        arn =_backup_table("backup")
    for s in secrets:
        i += 1
        domain = s.domain
        access = s.access
        try:
            message = f"[{i}/{ns}] - Reconfiguring the secret ({domain},{access})"
            logger.info(message)
            iou.my_output(message)
            secret = get_secret(domain, access, old_mem)
            insert_secret("I", f"{domain}{SEPARATOR}{access}", secret.user_id,  secret.password,  secret.info, old_mem, ekey)
            rename_secret(domain, access, "D", f"{domain}{SEPARATOR}{access}")
            rename_secret("I", f"{domain}{SEPARATOR}{access}", domain, access)
            delete_secret("D", f"{domain}{SEPARATOR}{access}")
        except Exception as e:
            logger.error(e)
            message = f"Could not reconfigure ({domain},{access})"
            iou.my_output(message)
            logger.error(message)

    return arn


def count_secrets():
    """Returns the total number of secrets"""
    filter_secret = Secret()
    return len(_get_table().query_record(filter_secret))





from datetime import datetime

import boto3
from boto3.dynamodb.conditions import Key

from ..constants import Secret
from ..utils.cryptutils import encrypt, encrypt_info
from .table import Table


MAX_REPEAT = 3
SUCCESS = 200

#some utility functions
def make_secret(d:dict)->Secret:
        return Secret(domain=d['domain'],
                      access=d['access'],
                      user_id = d['uid'],
                      password = d['pwd'],
                      info = d['info'],
                      timestamp = d['timestamp'])    

"""
    Proxy class for a table in a AWS DynamoDB database
    The methods can throw so the client (dbutils functions) need to capture them
"""
class AWSDynamoTable(Table):

    def __init__(self, table_name:str, profile_name:str):
        """
        Construct a table identified by a table name and a AWS profile_name
        """
        self.table_name = table_name
        self.profile_name = profile_name
        
    def get_table(self)->object:   
        session = boto3.session.Session(profile_name=self.profile_name)
        dynamodb = session.resource('dynamodb')
        return dynamodb.Table(self.table_name)
    
    def backup_table(self, backup_name:str):
        session = boto3.session.Session(profile_name=self.profile_name)
        client = session.client('dynamodb')
        res = client.create_backup(TableName = self.table_name, BackupName = backup_name)
        return res['BackupDetails']['BackupArn']

    def cleanup_table_backups(self, backup_name:str):
        session = boto3.session.Session(profile_name=self.profile_name)
        client = session.client('dynamodb')
        res = client.list_backups(TableName=self.table_name)
        for bkp in res['BackupSummaries']:
            if bkp['BackupName'] == backup_name:
                client.delete_backup(BackupArn=bkp['BackupArn'])

    def drop_table(self):
        session = boto3.session.Session(profile_name=self.profile_name)
        dynamodb = session.resource('dynamodb')
        dynamodb.Table(self.table_name).delete()
        
    def has_table(self, table_name:str=None)->bool:
        "Checks if the table exists"
        if table_name is None:
            table_name = self.table_name
        session = boto3.session.Session(profile_name=self.profile_name)
        dynamodb = session.resource('dynamodb')
        names = [x.table_name for x in dynamodb.tables.all()]
        return table_name in names
    
    def create_table(self, table_name:str=None):
        "Creates a table if it does not exist"
        if table_name is None:
            table_name = self.table_name        
        if self.has_table(table_name):
            return
        session = boto3.session.Session(profile_name=self.profile_name)
        dynamodb = session.resource('dynamodb')
        dynamodb.create_table(
            TableName=f"{table_name}",
            # Declare your Primary Key in the KeySchema argument
            KeySchema=[
                {
                    "KeyType": "HASH",
                    "AttributeName": "domain"
                    },
                {
                    "KeyType": "RANGE",
                    "AttributeName": "access"
                }
            ],
    
            # Any attributes used in KeySchema or Indexes must be declared in AttributeDefinitions
            AttributeDefinitions=[
                {
                    "AttributeName": "access",
                    "AttributeType": "S"
                },
                {
                    "AttributeName": "domain",
                    "AttributeType": "S"
                }
            ],
            # ProvisionedThroughput controls the amount of data you can read or write to DynamoDB per second.
            # You can control read and write capacity independently.
            ProvisionedThroughput={
                "ReadCapacityUnits": 5,
                "WriteCapacityUnits": 5
            },
        )
        dynamodb.Table(self.table_name).wait_until_exists()
        
    def insert_record(self,
                      secret:Secret,
                      mem_pwd:str,
                      salt:str,
                      timestamp:str = datetime.now().isoformat()) -> None:
        "insert a record in the secrets table"
        rep = 0
        status = 0
        #if the insert did not work, repeat up to MAX_REPEAT times
        while status != SUCCESS and rep < MAX_REPEAT:
            resp = self.get_table().put_item(Item={'domain'    : secret.domain,
                                                   'access'    : secret.access,
                                                   'uid'       : encrypt(secret.user_id, mem_pwd, salt),
                                                   'pwd'       : encrypt(secret.password, mem_pwd, salt),
                                                   'info'      : encrypt_info(secret.info, mem_pwd, salt),
                                                   'timestamp' : timestamp})
            status = resp['ResponseMetadata']['HTTPStatusCode']
            rep +=1

    def insert_encrypted_record(self,
                                secret:Secret,
                                timestamp:str = datetime.now().isoformat()) -> None:
        "insert a pre-encrypted record in the secrets table"
        rep = 0
        status = 0
        #if the insert did not work, repeat up to MAX_REPEAT times
        while status != SUCCESS and rep < MAX_REPEAT:        
            resp = self.get_table().put_item(Item={'domain'    : secret.domain,
                                                   'access'    : secret.access,
                                                   'uid'       : secret.user_id,
                                                   'pwd'       : secret.password,
                                                   'info'      : secret.info,
                                                   'timestamp' : timestamp})
            status = resp['ResponseMetadata']['HTTPStatusCode']
            rep +=1            
        
    def get_record(self,
                   secret: Secret) -> Secret:
        "retrieves an encrypted record keyed by domain and access, as a dictionary"
        rep = 0
        status = 0
        #if the insert did not work, repeat up to MAX_REPEAT times
        while status != SUCCESS and rep < MAX_REPEAT:        
            resp = self.get_table().get_item(Key={'domain'  : secret.domain,
                                                  'access'  : secret.access})
            status = resp['ResponseMetadata']['HTTPStatusCode']
            rep +=1            
            if 'Item' in resp:
                return make_secret(resp['Item'])
        return None
        
    
    def update_record_single_info(self,
                                  secret:Secret,
                                  mem_pwd:str,
                                  salt:str,
                                  timestamp:str = datetime.now().isoformat()) -> None:
        "update a record in the secrets table, with a maximum of one extra info stored in the Secret record"
        
        # #domain is the simbolyc field name and maps table column 'domain' to id #domain
        expression_attributes = {'#domain':'domain',
                                 '#access':'access'}
        # :domain is the symbolic value to be used in expressions for domain
        expression_values = {':domain':secret.domain,
                             ':access':secret.access}
    
        update_expression = "SET"
        condition_expression = "#domain = :domain AND #access = :access"

        if secret.user_id is not None:
            expression_attributes.update({'#uid':'uid'})
            expression_values.update({':uid' : encrypt(secret.user_id, mem_pwd, salt)})
            update_expression += ' #uid = :uid,'
        if secret.password is not None:
            expression_attributes.update({'#pwd':'pwd'})
            expression_values.update({':pwd' : encrypt(secret.password, mem_pwd, salt)})
            update_expression += ' #pwd = :pwd,'
        if secret.info_key is not None and secret.info_value is not None:
            expression_attributes.update({'#info':'info','#key':secret.info_key})
            expression_values.update({':info':encrypt(secret.info_value, mem_pwd, salt)})
            update_expression += ' #info.#key = :info,'
            
        #if nothing to update then return
        if update_expression == 'SET':
            return
        #now add the timestamp
        expression_attributes.update({'#timestamp':'timestamp'})
        expression_values.update({':ts': timestamp})
        update_expression += ' #timestamp = :ts'

        rep = 0
        status = 0
        #if the insert did not work, repeat up to MAX_REPEAT times
        while status != SUCCESS and rep < MAX_REPEAT:
            resp = self.get_table().update_item(Key={"domain": secret.domain, "access": secret.access},
                                                ExpressionAttributeNames  = expression_attributes,
                                                ExpressionAttributeValues = expression_values,
                                                UpdateExpression          = update_expression,
                                                ConditionExpression       = condition_expression
                                                )
            status = resp['ResponseMetadata']['HTTPStatusCode']
            rep +=1
        
    def update_record_info_dictionary(self,
                                      secret: Secret,
                                      timestamp:str = datetime.now().isoformat()) -> None:
        "update the info dictionary for a record in the secrets table"
        
        einfo = secret.encrypted_info
        if einfo is None:
            return
        
        expression_attributes = {'#domain':'domain',
                                 '#access':'access',
                                 '#info': 'info',
                                 '#timestamp':'timestamp'}
        expression_values = {':domain':secret.domain,
                             ':access':secret.access,
                             ':info': einfo,
                             ':ts': timestamp}
        update_expression = "SET #info = :info, #timestamp = :ts"
        condition_expression = "#domain = :domain AND #access = :access"

        rep = 0
        status = 0
        #if the insert did not work, repeat up to MAX_REPEAT times
        while status != SUCCESS and rep < MAX_REPEAT:        
            resp = self.get_table().update_item(Key={"domain": secret.domain, "access": secret.access},
                                                ExpressionAttributeNames  = expression_attributes,
                                                ExpressionAttributeValues = expression_values,
                                                UpdateExpression          = update_expression,
                                                ConditionExpression       = condition_expression
                                                )
            status = resp['ResponseMetadata']['HTTPStatusCode']
            rep +=1
        
    def delete_record(self, secret:Secret) -> None:
        "delete a single record based on domain and access keys"
        rep =0
        status =0
        #if the insert did not work, repeat up to MAX_REPEAT times 
        while status != SUCCESS and rep < MAX_REPEAT:       
            resp = self.get_table().delete_item(Key={'domain'  : secret.domain,
                                                     'access'  : secret.access})
            status = resp['ResponseMetadata']['HTTPStatusCode']
            rep +=1
        
    def query_record(self, secret:Secret) -> list: 
        "return a list of secrets matching the domain key passed"
        
        secrets = []
        if secret.domain is not None:
            resp = self.get_table().query(KeyConditionExpression=Key('domain').eq(secret.domain))
        else:
            resp = self.get_table().scan()
        for i in resp['Items']:
            secrets.append(Secret(domain=i['domain'],access=i['access']))
        #sort the list
        secrets.sort(key=lambda x: x.domain+x.access)

        return secrets        
        
                      
        



import json
import os
import pytest
from secretwallet.constants import parameters, Parameters, AWS_PROFILE, TEST_TABLE

@pytest.fixture
def set_up():
    path = os.path.dirname(__file__)
    conf_file = os.path.join(path,'data','.secretwallet','test.json')
    #overrides
    data_overrides = {'profile'            : 'my profile',
                      'pre_salt'           : 'my salt',
                      'table_name'         : 'my table'}
    
    
    os.makedirs(os.path.dirname(conf_file), exist_ok=True)
    with open(conf_file, 'w') as cfile:
        json.dump(data_overrides, cfile)    
    yield conf_file
    
    parameters.clear()
    if os.path.exists(conf_file):
        os.remove(conf_file)
        os.rmdir(os.path.dirname(conf_file))
        
        
        
def test_parameters_singleton():
    assert parameters is not None
    assert AWS_PROFILE == parameters.get_profile_name()
    assert TEST_TABLE == parameters.get_table_name()
    assert parameters is Parameters() #singleton test
    
def test_parameters_overrides():
    data_overrides = {'profile'    : 'my profile',
                      'pre_salt'   : 'my salt',
                      'table_name' : 'my table'}
    parameters.set_data(data_overrides)
    assert data_overrides['profile']    == parameters.get_profile_name()
    assert data_overrides['pre_salt']   == parameters.get_pre_salt()
    assert data_overrides['table_name'] == parameters.get_table_name()
    
def test_parameters_configure(set_up):
    conf_file = set_up
    parameters.configure(conf_file)
    assert 'my profile'             == parameters.get_profile_name()
    assert 'my salt'                == parameters.get_pre_salt()
    assert 'my table'               == parameters.get_table_name()    


import os

from secretwallet.constants import parameters, DB_AWS_DYNAMO, DB_LOCAL_SQLITE
from secretwallet.main.configuration import get_configuration
import secretwallet.utils.dbutils as du


path = os.path.dirname(__file__)
conf_file = os.path.join(path,'data','test_integration.json')
conf_data = get_configuration(conf_file)
parameters.set_data(conf_data)

#here create all test tables with different storage types and then reset to default

#local storage
parameters.set_storage_type(DB_LOCAL_SQLITE)
du.create_table(parameters.get_table_name())

#aws
parameters.set_storage_type(DB_AWS_DYNAMO)
du.create_table(parameters.get_table_name())


import getpass

from secretwallet.constants import parameters
from secretwallet.utils.logging import get_logger


logger = get_logger(__name__, parameters.get_log_level())
parameters.register_logger(__name__, logger)

def display_list(message, secrets):
    field_lenght = max([len(x.domain) for x in secrets], default=0)+5
    format_header = f"<%-{field_lenght-1}s: <access>"
    format_record = f"%-{field_lenght}s: %s"
    print("**********************************************************")
    print(f"{message}: ")
    print(format_header%'domain>')
    for s in secrets:
        print(format_record%(s.domain,s.access))
    print("**********************************************************")

def display_numbered_list(message, secrets):
    field_lenght = max([len(x[0]) for x in secrets], default=0)+5
    format_header = f"%-7s: <%-{field_lenght-1}s: <access>"
    format_record = f"%-7d: %-{field_lenght}s: %s"
    print("**********************************************************")
    print(f"{message}: ")
    print(format_header%('<num>','domain>'))
    idx = 1;
    for s in secrets:
        print(format_record%(idx, s.domain, s.access))
        idx += 1
    print("**********************************************************")

def get_secret_by_idx(message, secrets):
    display_numbered_list(message, secrets)
    while True:
        answ = my_input("\nWhich secret? Type the index number in the list above or 0 to quit:  ")
        try:
            idx = int(answ)
            if (idx==0):
                return None
            if idx > 0 and idx <= len(secrets):
                return secrets[idx-1]
            print("Index out of range. Please retry")
        except Exception as e:
            print(e)
            print("I need a number. Please retry")

def confirm_delete(secrets):
    "Confirm secrets to delete"
    display_list("Secrets to delete", secrets)
    answ = my_input("\nDo you want to delete these secrets (yes|no) ")
    if not answ.lower().startswith('y'):
        exit(1)

def confirm_delete_key(domain, access, key):
    "Confirm secrets to delete"
    answ = my_input(f"Do you want to delete the key {key} for domain {domain} and access {access}? (yes|no) ")
    if not answ.lower().startswith('y'):
        exit(1)

def confirm_rename(secrets):
    "Confirm secret to rename"
    display_list("Secret to rename", secrets)
    answ = my_input("\nDo you want to rename this secret (yes|no) ")
    if not answ.lower().startswith('y'):
        exit(1)

def my_input(question):
    "Mockable input function"
    return input(question)

def my_output(message, want_exit=False, with_logging = True):
    "Mockable output function"
    if with_logging:
        logger.info(message)
    print(message)
    if want_exit:
        exit(1)

def my_getpass(question):
    "Mockable getpass function"
    return getpass.getpass(question)

def my_parse(parser, args):
    try:
        return parser.parse_args(args)
    except SystemExit:
        if parameters.is_in_shell():
            parser.print_usage()
            return None
        else:
            raise

class MockableInput:
    def __init__(self,inputs) -> None:
        "Passes a sets of inputs to be replayed"

        #internal generator function
        def generator(inputs) -> str:
            while len(inputs) > 0:
                yield inputs.pop(0)
            return

        self.__generator = generator(inputs)

    def __call__(self, question:str) -> str:
        myinput = self.__generator.__next__()
        logger.debug(f"{question}: {input}")
        return myinput


def display_reconfiguration_warning():
    "Display a warning when reconfiguring the system"
    print("*******************************************************************")
    print("""
  You are performing a reconfiguration of the secrets' remote
  table. This is because you are either changing the memorable-
  or the device password. This will result in all of your
  secrets being re-encrypted with the new password. You will
  not be able to retrieve them with the old keys.
  This operation takes time, depending on how large is the
  table. Be patient!
  A backup of the original table is performed, in case you need
  to roll back the table.
  You can manage your time-stamped backups from the AWS
  Management console or using the secret-wallet command line.
          """)
    print("*******************************************************************")
    answ = input("\nDo you want to go ahead? (yes|no) ")
    if answ.lower().startswith('y'):
        return
    else:
        exit(1)


def display_secret(secret):
    "Print a secret in a fixed format"
    print("**********************************************************")
    print("Secret id:")
    print(f"domain              : {secret.domain}")
    print(f"access              : {secret.access}")
    print("\nSecret credentials:")
    print(f"login               : {secret.user_id}")
    print(f"password            : {secret.password}")
    if secret.info is not None:
        print("\nSecret extra info:")
        for k,v in secret.info.items():
            print(f"{k:20}: {v}")
    print(f"\nLast updated        : {secret.timestamp}")
    print("**********************************************************")

def display_all_secrets(secrets):
    "Return a text representation of all the secrets in a readable format"
    for secret in secrets:
        display_secret(secret)



import datetime
from multiprocessing import Process
from multiprocessing.connection import Client, Listener
from time import sleep

from secretwallet.constants import parameters
from secretwallet.utils.logging import get_logger
from secretwallet.utils import is_posix

logger = get_logger(__name__, parameters.get_log_level())
parameters.register_logger(__name__, logger)

def session_listener(seed, timeout):
    """ Session service function with an initial value as a seed
    input: seed    the initial value stored in the session
           timeout the validity period of the session value (in seconds
    """
    logger.info("Listener starts")
    serv = Listener(parameters.get_session_address(), authkey=parameters.get_session_connection_password())
    serv.password = seed
    serv.pwd_timeout = timeout
    serv.start_time = datetime.datetime.now()
    serv.origin_time = datetime.datetime.now()
    while True:
        with serv.accept() as conn:
            try:
                req = conn.recv()
            except EOFError:
                break
            if 'action' in req and req['action'] == 'set':
                serv.password = req['password']
                serv.start_time = datetime.datetime.now()
                conn.send({'status':'done','password':serv.password})
                req.clear()
            elif 'action' in req and req['action'] == 'get':
                if (datetime.datetime.now() - serv.start_time).total_seconds() < serv.pwd_timeout:
                    conn.send({'status':'fresh','password':serv.password})
                    serv.start_time = datetime.datetime.now()
                else:
                    serv.password =  None
                    conn.send({'status':'stale','password':None})
                req.clear()
            elif 'action' in req and req['action'] == 'stop':
                logger.debug("Goodbye from listener")
                conn.send({'status':'terminated','password':None})
                req.clear()
                break #break here to stop the server
            elif 'action' in req and req['action'] == 'ping':
                logger.debug("Ping request from listener")
                conn.send({'status':'OK','password':None})
                req.clear()
            else:
                conn.send({'status':'bad command','password':None})
                req.clear()
    logger.info("Listener ends")


def session_sweeper(lifetime):
    "The process that will kill the session daemon eventually"
    logger.info("sweeper starts")
    sleep(lifetime)
    conn = Client(parameters.get_session_address(), authkey=parameters.get_session_connection_password())
    conn.send({'action':'stop','password':None})
    ret = conn.recv()
    logger.debug(ret['status'])
    logger.info("Sweeper ends")

def my_session(value, lifetime, timeout):
    p = Process(target=session_sweeper, args=(lifetime,))
    q = Process(target=session_listener, args=(value, timeout))
    p.daemon = True
    q.daemon = True

    p.start()
    q.start()
    #keep track of process pids
    parameters.set_listener_pid(q.pid)
    parameters.set_sweeper_pid(p.pid)

    p.join()
    q.join()

    p.terminate()
    q.terminate()

def start_my_session(value, lifetime, timeout):
    if is_posix() and not parameters.is_in_shell():
        import daemon
        with daemon.DaemonContext():
            my_session(value, lifetime, timeout)
    else: #Windows or other system don't support daemons in the way required
        pass




'''
Created on 13 Jan 2020

@author: codimoc
'''

import os
import pytest
import secretwallet.main.configuration as mc

@pytest.fixture
def set_up():
    path = os.path.dirname(__file__)
    conf_file = os.path.join(path,'data','.secretwallet','credentials')
    yield conf_file
    
    if os.path.exists(conf_file):
        os.remove(conf_file)
        os.rmdir(os.path.dirname(conf_file))
        

def test_no_aws_configuration(set_up):
    conf_file = set_up
    assert False == mc.has_credentials(conf_file)
    
def test_create_and_check_aws_configuration(set_up):
    conf_file = set_up
    access_key = '1234'
    secret_access_key = 'Abc4'
    region = 'my region'
    mc.set_credentials(conf_file, access_key, secret_access_key, region)
    assert True == mc.has_credentials(conf_file)
    credentials = mc.get_credentials(conf_file)
    assert access_key        == credentials.get('aws_access_key_id')
    assert secret_access_key == credentials.get('aws_secret_access_key')
    assert region            == credentials.get('region')
    
def test_change_and_check_aws_configuration(set_up):
    conf_file = set_up
    access_key = '1234'
    secret_access_key = 'Abc4'
    region = 'my region'
    mc.set_credentials(conf_file, access_key, secret_access_key, region)
    assert True == mc.has_credentials(conf_file)
    
    access_key = '456'
    secret_access_key = 'three cars'
    region = 'your region'
    mc.set_credentials(conf_file, access_key, secret_access_key, region)    
    credentials = mc.get_credentials(conf_file)
    assert access_key        == credentials.get('aws_access_key_id')
    assert secret_access_key == credentials.get('aws_secret_access_key')
    assert region            == credentials.get('region')    


from collections import namedtuple
import datetime
import json
import logging
from os.path import expanduser, exists

from .utils import make_version_number


if 'HOME_DIR' in globals():
    HOME_FOLDER = eval('HOME_DIR')
else:
    HOME_FOLDER = expanduser("~")


CONFIG_FOLDER = f"{HOME_FOLDER}/.secretwallet"
CONFIG_FILE = f"{CONFIG_FOLDER}/secretwallet.json"
CREDENTIALS_FOLDER = f"{HOME_FOLDER}/.aws"
CREDENTIALS_FILE = f"{CREDENTIALS_FOLDER}/credentials"

PRE_SALT = b"Nel mezzo del cammin di nostra vita"

#dynamoDB variables
SECRET_ACCESS_TABLE='access_secrets'
TEST_TABLE = 'test'

#AWS configuration
AWS_PROFILE='secret-wallet'

#session parameters
SESSION_ADDRESS = ('localhost',6417)
SESSION_PWD = b'yooCani3'
SESSION_TIMEOUT  = 60  #number of second the mem password is kept fresh
SESSION_LIFETIME = 600 #lifetime in seconds of the entire session


#Password policy
PWD_ATTEMPTS = 6 #number of attempts to get a good password
PWD_LENGTH = 8 #at least 8 chars long
PWD_NUMBERS=1  #at least one number
PWD_SPECIAL=1  #at least one special char
PWD_UPPER = 1  #at least an upper case

LOG_FILE = f"{CONFIG_FOLDER}/secretwallet.log"
LOG_LEVEL = "info"
LOG_MAX_FILE_SIZE =  1000000 #1MB
LOG_BACKUP_COUNT  = 1        #number of rotated backup files that are retained

#type of storage
DB_AWS_DYNAMO = 'aws_dynamo'
DB_LOCAL_SQLITE = 'local_sqlite'


def make_log_level(level):
    if level.lower()=="critical":
        return logging.CRITICAL
    elif level.lower()=="fatal":
        return logging.FATAL
    elif level.lower()=="error":
        return logging.ERROR
    elif level.lower()=="warning":
        return logging.WARNING
    elif level.lower()=="info":
        return logging.INFO
    elif level.lower()=="debug":
        return logging.DEBUG
    else:
        return logging.NOTSET

#an object to store configurable parameters

def singleton(cls):
    instances = {}
    def wrapper(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return wrapper

@singleton
class Parameters(object):

    def __init__(self):
        self.__data = dict()
        self.__loggers = dict()

    def set_data(self,data):
        self.__data = dict(data)
        self.update_loggers()
        
    def get_data(self):
        return self.__data

    def register_logger(self, name, logger):
        self.__loggers[name] = logger

    def update_loggers(self):
        level = make_log_level(self.get_log_level())
        for logger in self.__loggers.values():
            logger.setLevel(level)
            handler = logging.handlers.RotatingFileHandler(LOG_FILE,
                                                           mode='a',
                                                           maxBytes=LOG_MAX_FILE_SIZE,
                                                           backupCount=LOG_BACKUP_COUNT,
                                                           encoding='utf-8',
                                                           delay=0)
            handler.setLevel(level)
            # Create a formatter.
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            # Add handler and formatter.
            handler.setFormatter(formatter)
            logger.addHandler(handler)


    def configure(self, conf_file):
        if not exists(conf_file):
            raise FileNotFoundError("Missing configuration file: run the init command")
        with open(conf_file, 'r') as cfile:
            self.set_data(json.load(cfile))

    def clear(self):
        self.__data = dict()

    def get_profile_name(self):
        if 'profile' in self.__data:
            return self.__data['profile']
        else:
            return AWS_PROFILE

    def get_log_level(self):
        if 'log_level' in self.__data:
            return self.__data['log_level']
        else:
            return LOG_LEVEL

    def set_profile_name(self, profile):
        self.__data['profile'] = profile

    def get_pre_salt(self):
        if 'pre_salt' in self.__data:
            return self.__data['pre_salt']
        else:
            return PRE_SALT

    def get_table_name(self):
        if 'table_name' in self.__data:
            return self.__data['table_name']
        else:
            return SECRET_ACCESS_TABLE
        
    def get_storage_type(self):
        if 'storage_type' in self.__data:
            return self.__data['storage_type']
        else:
            return DB_AWS_DYNAMO
        
    def set_storage_type(self, storage):
        self.__data['storage_type'] = storage        

    def set_table_name(self, table):
        self.__data['table_name'] = table

    def get_salt_key(self):
        if 'key' in self.__data:
            return self.__data['key']
        else:
            raise RuntimeError('The encrypted key for the salt was not found')

    def set_salt_key(self, key):
        self.__data['key'] = key

    def get_session_timeout(self):
        if 'session_timeout' in self.__data:
            return self.__data['session_timeout']
        else:
            return SESSION_TIMEOUT

    def set_session_timeout(self, timeout):
        self.__data['session_timeout'] = timeout

    def set_clock_start(self, start):
        "set the time from which password timeout starts counting down"
        self.__data['clock_start'] = start

    def get_clock_start(self):
        "get the time from which password timeout started counting down"
        if 'clock_start' in self.__data:
            return self.__data['clock_start']
        else:
            return None
    def set_in_shell(self, flag):
        "set a flag to say if in shell or not"
        self.__data['in_shell'] = flag

    def is_in_shell(self):
        "check if we are running inside a shell"
        if "in_shell" in self.__data:
            return self.__data['in_shell']
        else:
            return False

    def get_memorable_pwd(self):
        "get the memorable password if stored during in shell mode"
        if not self.is_in_shell():
            return None
        if 'memorable' in self.__data and 'clock_start' in self.__data:
            now = datetime.datetime.now()
            if (now - self.get_clock_start()).total_seconds() < self.get_session_timeout():
                self.set_clock_start(now)
                return self.__data['memorable']
            else:
                self.set_memorable_pwd(None)
                return None
        else:
            return None

    def set_memorable_pwd(self, memorable):
        "set the memorable password during shell mode"
        self.__data['memorable'] = memorable
        self.set_clock_start(datetime.datetime.now())


    def get_session_lifetime(self):
        if 'session_lifetime' in self.__data:
            return self.__data['session_lifetime']
        else:
            return SESSION_LIFETIME

    def set_session_lifetime(self, lifetime):
        self.__data['session_lifetime'] = lifetime

    def get_session_address(self):
        if 'session_address' in self.__data:
            return self.__data['session_address']
        else:
            return SESSION_ADDRESS

    def get_session_connection_password(self):
        if 'session_connection_password' in self.__data:
            return self.__data['session_connection_password']
        else:
            return SESSION_PWD

    def set_listener_pid(self,pid):
        self.__data["listener_pid"] = pid

    def get_listener_pid(self):
        if "listener_pid" in self.__data:
            return self.__data["listener_pid"]
        else:
            return None

    def set_sweeper_pid(self,pid):
        self.__data["sweeper_pid"] = pid

    def get_sweeper_pid(self):
        if "sweeper_pid" in self.__data:
            return self.__data["sweeper_pid"]
        else:
            return None


#single object
parameters = Parameters()

# a message class containing secret info
secret_fields = ['domain','access','user_id','password','info', 'encrypted_info','info_key','info_value','timestamp']

#conditional compilation on python version >= 3.7
if make_version_number() >= 3700:
    default_vals = [None]*len(secret_fields)
    Secret = namedtuple('Secret', secret_fields, defaults=default_vals) #this for versions >= 3.7
else:           
    Secret = namedtuple('Secret', secret_fields) #this for versions < 3.7
    default_vals = (None,)*len(Secret._fields)
    Secret.__new__.__defaults__ = default_vals #this is required for version < 3.7
    



import base64

from secretwallet.constants import parameters 
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def encrypt_key(config_password):
    """Produces an encripted and repeatible key
       combining the PRE_SALT string and a user password specific for 
       configuration generation
       input:
       config_password    a password used to make the unique configuration file (string repr)
       output:
       the encrypted key used for encrypting data (string repr)"""
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                     length=32,
                     salt=parameters.get_pre_salt(),
                     iterations=100000,
                     backend=default_backend()
                     )
    key = base64.urlsafe_b64encode(kdf.derive(config_password.encode("latin1")))
    return key.decode("latin1")
            
def _encrypt(secret, mem_pswd, salt):
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                     length=32,
                     salt=salt.encode('latin1'),
                     iterations=100000,
                     backend=default_backend()
                     )
    key = base64.urlsafe_b64encode(kdf.derive(mem_pswd.encode("latin1")))   
    f = Fernet(key)
    return f.encrypt(secret.encode("latin1"))
    
def encrypt(secret, mem_pswd, salt):
    """Encrypts a secretwallet using a fixed key and a memorable password
    input:
    secret       text to encrypt (unicode)
    mem_pswd     memorable password (unicode)
    salt         a string representation of the salt (optional)    
    output:
    The encrypted (byte string) value 
    """
    return _encrypt(secret, mem_pswd, salt).decode('latin1')

def _decrypt(secret, mem_pswd, salt):
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                     length=32,
                     salt=salt.encode('latin1'),
                     iterations=100000,
                     backend=default_backend()
                     )
    key = base64.urlsafe_b64encode(kdf.derive(mem_pswd.encode("latin1")))   
    f = Fernet(key)
    return f.decrypt(secret).decode("latin1")    
 
def decrypt(secret, mem_pswd, salt):
    """Decrypts a secretwallet using a fixed key and a memorable password
    input:
    secret       encrypted secret
    mem_pswd     memorable password (unicode)
    salt         a string representation of the salt (optional)    
    output:
    The decrypted secret 
    """
    return _decrypt(secret.encode('latin1'), mem_pswd, salt)

def encrypt_info(info,mem_pwd, salt):
    """Encrypts a a full dictionary of extra info
    input:
    info         a dictionary of extra information
    mem_pwd      memorable password (unicode)
    salt         a string representation of the salt (optional)    
    output:
    The encrypted dictionary of extra information 
    """
    einfo = {}
    if info is not None:
        for key, value in info.items():
            einfo[key] = encrypt(value, mem_pwd, salt) #in string format
    return einfo

def decrypt_info(info, mem_pwd, salt):
    """Decrypts a a full dictionary of extra info
    input:
    info         an encrypted dictionary of extra information
    mem_pwd      memorable password (unicode)
    salt         a string representation of the salt (optional)    
    output:
    The decrypted dictionary of extra information 
    """    
    dinfo = {}
    if info is not None:
        for key, value in info.items():
            dinfo[key] = decrypt(value, mem_pwd, salt) #from string format
    return dinfo



from secretwallet.utils.password_manager import explain

def test_short_password():
    pwd = 'mamma'
    assert 'password length' in explain(pwd)
    
def test_missing_numbers_in_password():
    pwd = 'Mamm@ePapp@'
    assert 'minimum number of numbers in password' in explain(pwd)    
    
def test_missing_uppercase_password():
    pwd = 'mamm@3papp@'
    assert 'minimum number of upper-case characters' in explain(pwd)
    
def test_missing_special_password():
    pwd = 'Mamma3pappa'
    assert 'minimum number of special characters' in explain(pwd)
    
def test_good_password():
    pwd = 'Mamm@3pappa'
    assert explain(pwd) is None    